/*
 * $RCSfile: ioctl.c,v $
 *
 * Copyright (c) 1992  Arizona Board of Regents
 *
 * $Revision: 1.1 $
 * $Date: 1995/10/26 19:15:20 $
 * $Author: ho $
 *
 * $Log: ioctl.c,v $
 * Revision 1.1  1995/10/26 19:15:20  ho
 * Initial revision
 *
 * Revision 1.2  1992/08/15  01:19:59  davidm
 * select() totally revised---didn't work when user passed fd_sets which
 * were smaller than sizeof(fd_set); it is common practice (?) to use
 * just an "int" instead of fd_set
 *
 * Support for signal handling added.  If a server call returns EINTR,
 * the library checks the BSD server for pending signals by executing
 * a NOP system call (sigblock(sigmask(SIGKILL))).
 *
 * Revision 1.1  1992/07/22  18:39:58  davidm
 * Initial revision
 *
 */
#include <sys/ioctl.h>
#include "xsi_main.h"
#include "prot/ip.h"
#include <net/route.h>

int
ioctl(int s, int request, char *argp)
{
    char *inp;
    int inlen;
    char *outp;
    int outlen;
    int result;

    xTrace3(xsi, TR_FUNCTIONAL_TRACE, "ioctl(s=%d,request=%d,argp=%08x)",
	    s, request, argp);

    if (!xsi_server) {
	xsi_user_init();
    } /* if */

    if (FD_ISSET(s, &is_xsi_fd)) {
        int copyoutlen = 0;
        char * copyoutaddr;
        char * copyoutbuf;

	if (request & IOC_IN) {
	    inp = argp;
	    inlen = IOCPARM_LEN(request);
	} /* if */
	if (request & IOC_OUT) {
	    outp = argp;
	    outlen = IOCPARM_LEN(request);
	} /* if */
        /* stupid place to put data translation */
        if (request == SIOCDELRT || request == SIOCADDRT) {
          IProuteInfo ipr_entry;
          route_BSD_2_Xk(&ipr_entry, argp);
          inp = (char *)&ipr_entry;
          inlen = sizeof(IProuteInfo);
        }
	result = XSI(ioctl, (xsi_server, xsi_cid, s, request,
			   inp, inlen, &outp, &outlen,
                           &copyoutaddr, &copyoutbuf, &copyoutlen, &stat));
        if (copyoutlen) {
          bcopy(copyoutbuf, copyoutaddr, copyoutlen);
        }
        return result;
    } else {
	return sc_ioctl(s, request, argp);
    } /* if */
} /* ioctl */

			/*** end of ioctl.c ***/
