/*
 * cookie.c
 *
 * x-kernel v3.2
 *
 */

/*
 *  The cookie protocol implements an idea proposed by Phil Karn
 *  to lessen the potential for "clogging" attacks; it enforces
 *  the provision that a remote participant must present a unique number
 *  assigned to it by the local host before communication can proceed.

 */

int tracecookie_photurisp;

#include "xkernel.h"
#include "md5hash.h"
#include "udp.h"  

#include "cookie_photuris_i.h"
#include <stdlib.h>
#include <memory.h>

#ifdef __STDC__

void			cookie_photuris_init( XObj );
static XObj		cookieOpen( XObj, XObj, XObj, Part * );
static xkern_return_t	cookieOpenDone( XObj, XObj, XObj, XObj );
/* static xkern_return_t	cookieOpenDisable( XObj, XObj, XObj, Part * );*/
static xkern_return_t	cookieOpenEnable( XObj, XObj, XObj, Part * );
static xmsg_handle_t	cookiePush( XObj, Msg * );
static xkern_return_t	cookiePop( XObj, XObj, Msg *, void * );
static xkern_return_t	cookieDemux( XObj, XObj, Msg * );
static int		cookieControlProtl( XObj, int, char *, int ) ;
static int		cookieControlSessn( XObj, int, char *, int ) ;
static xkern_return_t	cookieClose( XObj );
static int		xCopyParts(Part **, char **, Part *, XObj);
static void		cookieCreateSessn( XObj );
static cookie_t *	new_cookie(HOSTTYPE *, char *) ;
static long 		cookieTake(void *, char *, long, void *);
static void 		cookiePut(void *, char *, long, void *);
#else
void			cookie_photuris_init(  );
static XObj		cookieOpen();
static xkern_return_t	cookieOpenDone();
/* static xkern_return_t	cookieOpenDisable( );*/
static xkern_return_t	cookieOpenEnable(  );
static xmsg_handle_t	cookiePush(  );
static xkern_return_t	cookiePop( );
static xkern_return_t	cookieDemux(  );
static int		cookieControlProtl(  ) ;
static int		cookieControlSessn(  ) ;
static xkern_return_t	cookieClose( );
static int		xCopyParts();
static void		cookieCreateSessn( );
static cookie_t *	new_cookie() ;
static long 		cookieTake();
static void 		cookiePut();

#endif

#define xGetParticipants(lls, p, l) \
  (xControl(lls, GETPARTICIPANTS, (char *)p, l))



static int
xCopyParts(q, qbuf, p, lls)
     Part **q;
     char **qbuf;
     Part *p;
     XObj lls;
{
  char *ebuf;
  int	len = 500;

  ebuf = xMalloc(500);
  if (lls != ERR_XOBJ) {
    if (xGetParticipants(lls, ebuf, 500)<0) return 0;
  }
  else
    partExternalize(p, ebuf, &len);
  *q = (Part *)xMalloc(partExtLen(ebuf)*sizeof(Part));
  partInternalize(*q, ebuf);
  *qbuf = ebuf;
  return partLen(((Part *)(*q)));
}

static XObj cookieOpen(self, hlp, hlpType, part)
     XObj self, hlp, hlpType;
     Part *part;
{
  XObj lls, sself = ERR_XOBJ;
  PState *ps = (PState *)(self->state);
  Part *partrem;
  HOSTTYPE *host;
  int len;
  char *ebuf = 0;
  MapKey  k;

  len = xCopyParts(&partrem, &ebuf, part, ERR_XOBJ);
  if (len >= REMOTE_PART) host = (HOSTTYPE *)partPop(partrem[REMOTE_PART]);
  

  if ((lls = xOpen(self, hlpType, xGetDown(self, 0), part)) != ERR_XOBJ)
    {
      	SState *ss;
    	int random;
	xControl(lls,UDP_ENABLE_CHECKSUM,NULL,0);
	sself = xCreateSessn(cookieCreateSessn, hlp, hlpType, self, 1, &lls);
	ss  = (SState *)sself->state;
	random = rand();
	ss->Initiator_cookie = new_cookie(host,(char *) &random );
	ss->Responder_cookie = NULL;
	ss->remhost = *host;
	ss->state = RESPONDER_COOKIE_MISSING;
	k.lls = lls;
	memcpy(k.Initiator_cookie,ss->Initiator_cookie,HASH_SIZE_BYTES); 
	memset(k.Responder_cookie,0,HASH_SIZE_BYTES);
	sself->binding = mapBind(ps->cookieMap, &k, sself);
      	if (ebuf) { xFree(ebuf); xFree((char *)partrem); }
      	return sself;
    }
  if (ebuf) { xFree(ebuf); xFree((char *)partrem); }
  return ERR_XOBJ;
}

static void
cookieCreateSessn( self )
    XObj	self;
{
  SState *ss;

  xTrace1(cookie_photurisp, TR_EVENTS, "cookieCreateSessn entered: args %x", self);

  ss = X_NEW(SState);
  self->state = ss;
  self->push = cookiePush;
  self->pop = cookiePop;
  self->control = cookieControlSessn;
  self->close = cookieClose;
}

static xkern_return_t cookieOpenEnable(self, hlp, hlpType, p)
     XObj self, hlp, hlpType;
     Part *p;
{
    
  ((PState *)self->myprotl->state)->hlp = hlp;
  ((PState *)self->myprotl->state)->hlpType = hlpType;
  return xOpenEnable(self, hlpType, xGetDown(self, 0), p);
}

static xkern_return_t	cookieOpenDone( self, lls, llp, hlpType  )
     XObj self, lls, llp, hlpType;
{

 
  xTrace1(cookie_photurisp, TR_EVENTS, "cookieOpenDone finished: self %x", self);
  return XK_SUCCESS;
}


/*
 *
 *   cookiePush
 *
 *
 */
static xmsg_handle_t cookiePush(self, msg)
     XObj self;
     Msg *msg;
{
  SState *ss = (SState *)self->state;
  switch ( ss->state )
  {
	case RESPONDER_COOKIE_MISSING:
		{
		cookie_t co;
		memset(co,0,HASH_SIZE_BYTES);
  		msgPush(msg, cookiePut, co, HASH_SIZE_BYTES, NULL);
 		msgPush(msg, cookiePut, ss->Initiator_cookie, HASH_SIZE_BYTES, NULL);
		break;
		}
 	case OPERATIONAL:
  		msgPush(msg, cookiePut, ss->Responder_cookie, HASH_SIZE_BYTES, NULL);
  		msgPush(msg, cookiePut, ss->Initiator_cookie, HASH_SIZE_BYTES, NULL);
		break;
	default:
		return XK_FAILURE;
  }

  return xPush(xGetDown(self, 0), msg);
}




static void cookiePut(src, dst, len, sst)
    void *src;
    char *dst;
    long len;
    void *sst;
{
  bcopy(src, dst, len);
}

static long cookieTake(dst, src, count, type)
    void * dst;
    char * src;
    long count;
    void * type;
{
	bcopy(src,dst,count);
	return count;
}



/*
 *  cookieDemux
 *
 */
static xkern_return_t cookieDemux(self, lls, msg)
     XObj self, lls;
     Msg  *msg;
{
  XObj s; 
  MapKey  k;
  unsigned char * t;
  Part *parts;
  HOSTTYPE *host;
  SState * ss;
  XObj sessn;
  char *ebuf = 0;
  PState *ps = ((PState *)self->myprotl->state);
  cookie_t gi, gr;
  xkern_return_t ret;

  msgPop(msg, cookieTake, gi, HASH_SIZE_BYTES, NULL);
  msgPop(msg, cookieTake, gr, HASH_SIZE_BYTES, NULL);
  t = msgPeek(msg,1,FALSE);
  k.lls = lls;
  memcpy(k.Initiator_cookie,gi,HASH_SIZE_BYTES);
  memcpy(k.Responder_cookie,gr,HASH_SIZE_BYTES);
  xTrace1(cookie_photurisp, TR_EVENTS, "cookieDemux entered: lls %x", lls);
  if ( mapResolve(ps->cookieMap, &k, &s) == XK_FAILURE ) 
  { 
	memset(k.Responder_cookie,0,HASH_SIZE_BYTES);
  	if ( mapResolve(ps->cookieMap, &k, &s) == XK_FAILURE ) 
  	{ 
		if  (*t == 0)  
			/* only accept cookie requests and exchange request */
		{
  			xCopyParts(&parts, &ebuf, 0, lls);
  			host = partPop(parts[REMOTE_PART]);
    			sessn = xCreateSessn(cookieCreateSessn, ps->hlp,
	 		ps->hlpType, self,1, &lls);
  			ss  = (SState *)sessn->state;
			ss->Initiator_cookie = X_NEW(cookie_t);
			memcpy(ss->Initiator_cookie,gi,HASH_SIZE_BYTES);
 			ss->Responder_cookie = new_cookie(host, ps->random);
  			ss->remhost = *host;
    			ss->state = OPERATIONAL;
  			k.lls = lls;
   			memset(k.Responder_cookie,0,HASH_SIZE_BYTES); 
  			memcpy(k.Initiator_cookie,ss->Initiator_cookie,HASH_SIZE_BYTES); 
  			if (ERR_BIND == (sessn->binding = mapBind(ps->cookieMap, &k, sessn)))
				return XK_FAILURE; /* not a problem since cookie request will
					be honored in photuris or the session will be closed 
					by photuris */
  			xDuplicate(lls);
  			xControl(lls,UDP_ENABLE_CHECKSUM,NULL,0);
  			if (ebuf) { xFree((char *)parts); xFree(ebuf); }
  			xOpenDone(ps->hlp, sessn, self);
			ret = xPop(sessn, lls, msg,0);
		}		
		else if (*t == 2)
		{
  			xCopyParts(&parts, &ebuf, 0, lls);
  			host = partPop(parts[REMOTE_PART]);
  			sessn = xCreateSessn(cookieCreateSessn, ps->hlp,
	 		ps->hlpType, self,1, &lls);
  			ss  = (SState *)sessn->state;
			ss->Initiator_cookie = X_NEW(cookie_t);
			memcpy(ss->Initiator_cookie,gi,HASH_SIZE_BYTES);
 			ss->Responder_cookie = new_cookie(host, ps->random);
			if ( memcmp(gr, ss->Responder_cookie, HASH_SIZE_BYTES) != 0)
				return XK_FAILURE; /* no valid cookie */
  			ss->remhost = *host;
    			ss->state = OPERATIONAL;
  			k.lls = lls;
   			memcpy(k.Responder_cookie,ss->Responder_cookie,HASH_SIZE_BYTES); 
  			memcpy(k.Initiator_cookie,ss->Initiator_cookie,HASH_SIZE_BYTES); 
  			if (ERR_BIND == (sessn->binding = mapBind(ps->cookieMap, &k, sessn)))
				return XK_FAILURE; /* there schouldn't be another session
					with the same ports/cookies */
  			xDuplicate(lls);
 		        xTrace1(cookie_photurisp, TR_EVENTS, "cookieOpenDone finished: self %x", self);
  			xControl(lls,UDP_ENABLE_CHECKSUM,NULL,0);
  			if (ebuf) { xFree((char *)parts); xFree(ebuf); }
  			xOpenDone(ps->hlp, sessn, self);
			ret = xPop(sessn, lls, msg,0);
		} else
		{
			/* responder got illegal message  */
  			xTrace0(cookie_photurisp, TR_EVENTS,
				 "cookieDemux: illegal message for uninitialized session");
			return XK_FAILURE;
		}
	} else
	{
  		ss = (SState *)s->state;
		if ( *t == 0)
			ret = xPop(s, lls, msg,0);
		else if ( *t == 1 )
		{
			/* intiator got message 1 */
			ss->Responder_cookie = X_NEW(cookie_t);
			memcpy(ss->Responder_cookie,gr,HASH_SIZE_BYTES);
			ss->state = OPERATIONAL;
  			mapRemoveBinding(ps->cookieMap,s->binding);
 			k.lls = lls;
  			memcpy(k.Responder_cookie,ss->Responder_cookie,HASH_SIZE_BYTES); 
  			memcpy(k.Initiator_cookie,ss->Initiator_cookie,HASH_SIZE_BYTES); 
  			s->binding = mapBind(ps->cookieMap, &k, s);
			ret = xPop(s, lls, msg,0);
		}
		else
			return XK_FAILURE;
	}
  }
  else
	ret = xPop(s, lls, msg,0);
  xTrace1(cookie_photurisp, TR_EVENTS, "cookieDemux demux to %x", s);
  return ret;
}


static  xkern_return_t
cookiePop ( self, lls, msg, hdr)
     XObj self, lls;
     Msg *msg;
     void *hdr;
{
   return xDemux(self, msg);
}

static int
cookieControlSessn(s, opcode, buf, len)
    XObj s;
    int opcode;
    char *buf;
    int len;
{
   SState	*ss = (SState *)s->state;
   switch (opcode) {
	case COOKIE_PHOTURIS_INITIATOR_COOKIE:
		{
    			if ((len != HASH_SIZE_BYTES) || ( ss->state == INITIATOR_COOKIE_MISSING))
				return XK_FAILURE;
			memcpy(buf,(char *) ss->Initiator_cookie,HASH_SIZE_BYTES);
			return XK_SUCCESS;
		}
		break;
	case COOKIE_PHOTURIS_RESPONDER_COOKIE:
		{
    			if ((len != HASH_SIZE_BYTES) || ( ss->state == RESPONDER_COOKIE_MISSING))
				return XK_FAILURE;
			memcpy(buf,(char *) ss->Responder_cookie,HASH_SIZE_BYTES);
 			return XK_SUCCESS;
		}
		break;	
      	default:
    		return xControl(xGetDown(s, 0), opcode, buf, len);
    }
}

static int
cookieControlProtl(s, opcode, buf, len)
    XObj s;
    int opcode;
    char *buf;
    int len;
{
   PState	*ps = (PState *)s->state;
   switch (opcode) {
	case COOKIE_PHOTURIS_SET_RANDOM_VALUE:
		{
    			if (len != RANDOM_SIZE)
				return XK_FAILURE;
			memcpy((char *)ps->random,buf,RANDOM_SIZE);
			return XK_SUCCESS;
		}
		break;
      	default:
    		return xControl(xGetDown(s, 0), opcode, buf, len);
    }   
}



static xkern_return_t	cookieClose( self )
  XObj self;
{
  SState * ss = (SState *) self->state; 
  XObj lls;
  lls = xGetDown(self,0);
  if (ss->Initiator_cookie != NULL)
              xFree((char *) ss->Initiator_cookie);
  if (ss->Responder_cookie != NULL)
              xFree((char *) ss->Responder_cookie);
  xFree(self->state);
  mapRemoveBinding( ((PState*)self->myprotl->state)->cookieMap,self->binding);
  xClose(lls);
  xDestroy(self);
  return XK_SUCCESS;
 
}


/*
 *  allocate and compute a cookie unique to the remote host
 *
 *
 */
static cookie_t *new_cookie(host, random)
  HOSTTYPE *host;
  char * random;
{
  MD5_CTX md5ctx;
  cookie_t *cookie = X_NEW(cookie_t);

  MD5Init(&md5ctx);
  MD5Update(&md5ctx, (unsigned char *)random, (unsigned int) RANDOM_SIZE);
  MD5Update(&md5ctx, (char *)host, sizeof(host));
  MD5Final(&md5ctx);
  bcopy(md5ctx.digest, (char *)cookie, HASH_SIZE_BYTES);
  return cookie;
}




void cookie_photuris_init( self )
     XObj self;
{
  PState *ps;

  xTrace0(cookie_photurisp, TR_GROSS_EVENTS, "cookie initializes");

  self->state = ps = X_NEW(PState);
  ps->cookieMap = mapCreate(COOKIE_MAP_SIZE, sizeof(MapKey));
  self->open = cookieOpen;
  self->opendone = cookieOpenDone;
  self->openenable = cookieOpenEnable;
  self->pop = cookiePop;
  self->push = cookiePush;
  self->demux = cookieDemux;
  self->control = cookieControlProtl;
}

