
/*   Byte order note:  Only Big Endian testing so far.
   No provision for defending against the "man in the middle" attack.
     This requires adding signatures.
 */

/* DHECKX misc design notes:

   RVRs mods to event code require "returning" semaphores.  I'm ignoring this.

   I decided not to do retries for the present.  If the offer message is lost,
   the code will time out after ten seconds and the Open will return failure.
   If I were to do retries, I should remember recent offers, in case an old
   one is accepted.

   If an Open is attempted on a session that is being set up, it fails.
   Arguably it should be queued.

   OpenDone at higher level can't send any messages until KX has processed
   first popped message with KX offer.

   Certain kinds of surprize Close are impossible because noone has a
   pointer to the current session.

   Future versions might provide for use of primes > 512 bits.
*/


/* #define TIME */  /* Turns on timing code.  Prints times to do key exchange. */

#include "xkernel.h"
#include "dheckxspi.h"
#include "saidmgr.h"
#include "dheckxspi_i.h"
#include "eth.h"		/* So ARP can go on top of DHECKX */
#include "ip.h"                 /* For IP pseudoheader fixup. */
#include "km.h"

extern int tracedheckxspip; int tracedheckxspip;


#ifdef __STDC__

static XObj		dheckxCreateSessn( XObj, XObj, XObj, XObj, int);
static void		dheckxSessnInit( XObj );
static xkern_return_t	dheckxDemux( XObj, XObj, Msg * );
static XObj		dheckxOpen( XObj, XObj, XObj, Part * );
static xkern_return_t	dheckxOpenEnable( XObj, XObj, XObj, Part * );
static xkern_return_t	dheckxOpenDisable( XObj, XObj, XObj, Part * );
static xkern_return_t	dheckxOpenDone( XObj, XObj, XObj, XObj);
static xkern_return_t   dheckxPop( XObj, XObj, Msg *, void * );
static xkern_return_t	dheckxClose( XObj );
static void		dheckxKeyExchangeInitiate(XObj);
static xkern_return_t	dheckxKeyExchangeReply(XObj, Msg *);
static xkern_return_t	dheckxKeyExchangeDone(XObj, Msg *);
static void		dheckxOpenWakeup(Event, void *);
/* int			xkrandom(unsigned char *, long); */
static long		dheckxcopy(void *, char *, long, void *);
static long		dheckxstore( char *, void *, long, void *);
void			dheckxenterkm(XObj, char*);

#else

static XObj		dheckxCreateSessn();
static void		dheckxSessnInit();
static xkern_return_t	dheckxDemux();
static XObj		dheckxOpen();
static xkern_return_t	dheckxOpenEnable();
static xkern_return_t	dheckxOpenDisable();
static xkern_return_t	dheckxOpenDone();
static xkern_return_t   dheckxPop();
static xkern_return_t	dheckxClose();
static void		dheckxKeyExchangeInitiate();
static xkern_return_t	dheckxKeyExchangeReply();
static xkern_return_t	dheckxKeyExchangeDone();
static void		dheckxOpenWakeup();
/* int			xkrandom(); */
static long		dheckxcopy();
static long		dheckxstore();
void			dheckxenterkm();

#endif /* __STDC__ */

#ifdef TIME
static void subtime(t1, t2, t3)    XTime *t1, *t2, *t3;
{   t3->sec = t2->sec - t1->sec;
    t3->usec = t2->usec - t1->usec;
    if (t3->usec < 0) { t3->usec += 1000000; t3->sec -= 1; } }
#endif

void	dheckxenterkm(XObj s, char key[SAIDMGR_KEYSIZE])
{
	KMB_MAPPING kmm;
        XObj	    kms;
        SState *    ss;
	Part	    kmp[1];

    	ss = (SState *)s->state;
	xControl(xGetDown(s,0), GETMYHOST, (char*) &kmm.local_host, sizeof(IPhost));
	kmm.local_algorithm = ss->algorithm;
	kmm.local_parameter = 0;
	kmm.local_mode	    = ss->mode;
	kmm.local_timeout   = -1;
	bcopy((char *)key, (char *)kmm.local_key, SAIDMGR_KEYSIZE);
	kmm.remote_said	    = ss->rsaid;
	xControl(xGetDown(s,0), GETPEERHOST, (char*) &kmm.remote_host, sizeof(IPhost));
	kmm.remote_algorithm = ss->algorithm;
	kmm.remote_parameter = 0;
	kmm.remote_mode      = ss->mode;
	kmm.remote_timeout   = -1;
	bcopy((char *)key, (char *)kmm.remote_key, SAIDMGR_KEYSIZE);

	xTrace3(dheckxspip, TR_MAJOR_EVENTS, "dheckxenetrekm %u %u %u ",ss->lsaid,
		 ss->rsaid, ss->algorithm  );


	partInit(kmp,1);
     	partPush(kmp[0],&ss->lsaid,sizeof(XSaid));	
     	kms = xOpen(s->myprotl, s->myprotl, xGetDown(s->myprotl,1) , kmp);
        if (kms == ERR_XOBJ) return;
     	xControl(kms, KM_SET, (char *) &kmm, sizeof(KMB_MAPPING));
     	xClose(kms); 
	
}
/* Initiate a key exchange.
   Called from Open, with the session (in state DHECKXINIT) as an argument.
   Composes a key exchange initiation message, and ships it.
   Sets session state to DHECKXINITWAIT to indicate we are waiting for a reply
   message to complete the exchange. */
static void dheckxKeyExchangeInitiate(sessn) XObj sessn;
{ SState *ss;
  unsigned long msgbuf[1+1+(2*DEGW)];
  unsigned long expo[DEGW] = FIVE( 0, 0, 0, 36264332u, 1256789 );
  Msg msg;

  xTrace1(dheckxspip, TR_MAJOR_EVENTS, "dheckxKeyExchangeInitiate entered: session = %x", sessn);
  xAssert(xIsSession(sessn));
  ss = sessn->state;
  xTrace2(dheckxspip, TR_MORE_EVENTS, "dheckxKeyExchangeInitiate: ss = %x, state = %d",
	  ss, ss->state);
  xAssert(ss->state == DHECKXINIT);

#ifdef TIME
  xGetTime(&ss->starttime);
#endif

  /* Get 512 random bits. */
  /*if (xkrandom((unsigned char *)xbuf, 16 * sizeof(long))) printf("xkrandom failed"); */

  memcpy(ss->k,expo,DEGW * sizeof(unsigned long));  

  bzero((char *)expo,DEGW * sizeof(unsigned long));

  xIfTrace(dheckxspip, TR_DETAILED)
  { int i; printf("dheckxKeyExchangeInitiate:\nexpo =");
   for (i=0;i<DEGW;i++) printf("%x:",ss->k[i]); printf("\n"); };

  DheckxKnown(ss->k,ss->kpx,ss->kpy);

  xIfTrace(dheckxspip, TR_DETAILED)
  { int i; printf("dheckxKeyExchangeInitiate: k*p = "); 
  for (i=0;i<DEGW;i++) printf("%x:",ss->kpx[i]);
  for (i=0;i<DEGW;i++) printf("%x:",ss->kpy[i]); printf("\n"); };

  /* build message with DHECKXMAGIC, state DHECKXINIT, gx padded to 512 bits. */
  msgbuf[0] = DHECKXMAGIC; msgbuf[1] = ss->state;

  memcpy(&msgbuf[2],ss->kpx,DEGW * sizeof(unsigned long));
  memcpy(&msgbuf[2+DEGW],ss->kpy,DEGW * sizeof(unsigned long));

  xIfTrace(dheckxspip, TR_DETAILED)
  { int i; printf("dheckxKeyExchangeInitiate: msg is ");
   for (i=0;i<1+1+(2*DEGW);i++)
     printf("%x:",msgbuf[i]); printf("\n"); };

  /* ship it, destroy message */
  msgConstructBuffer(&msg,(char *)msgbuf, (1+1+2*DEGW) * sizeof(unsigned long));
  msgPush(&msg, (MStoreFun) dheckxstore, &ss->lsaid, sizeof(ss->lsaid), NULL);
  msgPush(&msg, (MStoreFun) dheckxstore, &ss->mode, sizeof(ss->mode), NULL);
  msgPush(&msg, (MStoreFun) dheckxstore, &ss->algorithm, sizeof(ss->algorithm), NULL);
  msgPush(&msg, (MStoreFun) dheckxstore, &ss->rsaid, sizeof(ss->rsaid), NULL);
  xPush(xGetDown(sessn,TRANSPORT_PROTOCOL),&msg);
  msgDestroy(&msg);

  ss->state = DHECKXINITWAIT;  /* set state to DHECKXINITWAIT */
  xTrace1(dheckxspip, TR_EVENTS, "dheckxKeyExchangeInitiate exit: state = %d", ss->state);
}


/* Receive a key exchange initiation message, and reply.
   Called from Demux, with the session (in state DHECKXRESPONDER) and setup message
   as arguments.
   Computes the session key based on the setup message and our random data;
   compose and send an appropriate reply to the initiator, for him to
   complete his side of the key setup.
   Enter the computed key in the KM sessions.
   Set session state to DHECKXTRANSPARENT, putting us in transparent mode. */
static xkern_return_t dheckxKeyExchangeReply(sessn,msg) XObj sessn; Msg *msg;
{ SState *ss;
  unsigned long msgbuf[1+1+(2*DEGW)];
  unsigned long expo[DEGW] = FIVE( 0, 0, 0, 36264832u, 1234789 );
  unsigned char key[SAIDMGR_KEYSIZE];
  int t,tmp; Msg rmsg; Msg umsg;

  xTrace2(dheckxspip, TR_MAJOR_EVENTS,
	  "dheckxKeyExchangeReply entered: session = %x, msg = %x", sessn, msg);
  xAssert(xIsSession(sessn));
  ss = sessn->state;
  xTrace2(dheckxspip, TR_MORE_EVENTS, "dheckxKeyExchangeReply: ss = %x, state = %d",
	  ss, ss->state);
  xAssert(ss->state == DHECKXRESPONDER);

#ifdef TIME
  xGetTime(&ss->starttime);
#endif

  /* Check message length; unpack it. */
  xTrace1(dheckxspip, TR_DETAILED, "dheckxKeyExchangeReply: msgLen = %d", msgLen(msg));
  if ((tmp = msgLen(msg)) != sizeof(unsigned long) * (1+1+(2*DEGW)))
  { xTrace4(dheckxspip, TR_GROSS_EVENTS,
	"dheckxKeyExchangeReply: Session %x ss %x: Offer message %x length %d is wrong",
	    sessn, ss, msg, tmp);
    return XK_FAILURE; };
  msgPop(msg, dheckxcopy, (char *) msgbuf, (1+1+(2*DEGW))*sizeof(unsigned long), 0);

  xIfTrace(dheckxspip, TR_DETAILED)
  {  int i; printf("dheckxKeyExchangeInitiate: msg is ");
   for (i=0;i<1+1+(2*DEGW);i++)
     printf("%x:",msgbuf[i]); printf("\n");};

  /* Check msg starts with DHECKXMAGIC, and state of sender is DHECKXINIT. */
  if (!(msgbuf[0] == DHECKXMAGIC && msgbuf[1] == DHECKXINIT))
  { xTrace5(dheckxspip, TR_GROSS_EVENTS,
	    "dheckxKeyExchangeReply: Session %x ss %x: Offer message %x."
	    " Magic %d or SendState %d is wrong",
	    sessn, ss, msg, msgbuf[0], msgbuf[1]);
    return XK_FAILURE; };

  memcpy(ss->kpx,&msgbuf[2],DEGW * sizeof(unsigned long));
  memcpy(ss->kpy,&msgbuf[2+DEGW],DEGW * sizeof(unsigned long));


  /* Call xkrandom to get y, 512 bits. */
  /* $$ xkrandom is temporarily turned off. */
  /* if (xkrandom((unsigned char *)ybuf, 16 * sizeof(long))) printf("xkrandom failed"); */

  memcpy(ss->k,expo,DEGW * sizeof(unsigned long));  

  bzero((char *)expo,DEGW * sizeof(unsigned long));

  xIfTrace(dheckxspip, TR_DETAILED)
  {int i; printf("dheckxKeyExchangeInitiate:\nexpo =");
   for (i=0;i<DEGW;i++) printf("%x:",ss->k[i]); printf("\n"); };

  DheckxKnown(ss->k,ss->jpx,ss->jpy);


  xIfTrace(dheckxspip, TR_DETAILED)
  { int i; printf("dheckxKeyExchangeInitatiate: j*p = "); 
  for (i=0;i<DEGW;i++) printf("%x:",ss->jpx[i]);
  for (i=0;i<DEGW;i++) printf("%x:",ss->jpy[i]); printf("\n"); };

  ss->state = DHECKXREPLYING;  /* Advance session state to indicate we are replying. */

  /* reply message: DHECKXMAGIC, state = DHECKXREPLYING, gx & gy each padded to 512 */
  /* DHECKXMAGIC & gx are already present in msgbuf */

  msgbuf[1] = ss->state;

  memcpy(&msgbuf[2],&ss->jpx,DEGW * sizeof(unsigned long));
  memcpy(&msgbuf[2+DEGW],&ss->jpy,DEGW * sizeof(unsigned long));

  xIfTrace(dheckxspip, TR_DETAILED)
  {int i; printf("dheckxKeyExchangeInitiate: msg is ");
   for (i=0;i<1+1+(2*DEGW);i++)
     printf("%x:",msgbuf[i]); printf("\n");};

  /* Ship the reply.
  $$ Timing glitch: xPush releases the master lock.
     I could get another msg before I've finished computing and initialized
     the encryption keys.
     As long as the encryption protocol is above me in the graph, I can just
     throw away messages that arrive before I've finished computing the key.
     Presumably the correct thing to do is queue them. */

  msgConstructBuffer(&rmsg,(char *)msgbuf,(1+1+2*DEGW) * sizeof(unsigned long));
  msgPush(&rmsg, (MStoreFun) dheckxstore, &ss->lsaid, sizeof(ss->lsaid), NULL);
  msgPush(&rmsg, (MStoreFun) dheckxstore, &ss->mode, sizeof(ss->mode), NULL);
  msgPush(&rmsg, (MStoreFun) dheckxstore, &ss->algorithm, sizeof(ss->algorithm), NULL);
  msgPush(&rmsg, (MStoreFun) dheckxstore, &ss->rsaid, sizeof(ss->rsaid), NULL);
 
  /* push delayed to quarantie entry in key manager first */

#ifdef TIME
  xGetTime(&ss->endtime);
#endif

  Dheckx(ss->kpx,ss->kpy,ss->k,ss->jpx,ss->jpy);

  memcpy(key,ss->jpx,(DEGW-1)*sizeof(int));
  memcpy(&key[(DEGW-1)*sizeof(int)],ss->jpy,(DEGW-1)*sizeof(int));

  dheckxenterkm(sessn, key);

  xIfTrace(dheckxspip, TR_DETAILED)
  { int i; printf("dheckxKeyFinalKey: k*j*p = "); 
  for (i=0;i<DEGW;i++) printf("%x:",ss->jpx[i]);
  for (i=0;i<DEGW;i++) printf("%x:",ss->jpy[i]); printf("\n"); };

  bzero((char *)key,4*sizeof(int)); t=0;  /* Zero out key array on stack. */
  /* At this point, I don't really need to remember anything: the key has been
     handed off to the key manager, and all my intermediate values can be trashed. */

  ss->state = DHECKXTRANSPARENT;  /* go transparent */

#ifdef TIME
  xGetTime(&ss->endtime2);
  subtime(&ss->starttime,&ss->endtime,&ss->deltatime);
  subtime(&ss->starttime,&ss->endtime2,&ss->deltatime2);
  printf("\nResponder replied after %d.%06d seconds, and finished key calculation after %d.%06d seconds.\n",
     ss->deltatime.sec, ss->deltatime.usec, ss->deltatime2.sec, ss->deltatime2.usec);
#endif
  
  msgConstructBuffer(&umsg,(char *)&ss->lsaid,sizeof(XSaid));
  xDemux(sessn , &umsg );
  msgDestroy(&umsg);  /* destroy message */

  xPush(xGetDown(sessn,TRANSPORT_PROTOCOL),&rmsg); /* delayed to guarantie entry in keymanager*/
  msgDestroy(&rmsg);  /* destroy message */

  xTrace1(dheckxspip, TR_EVENTS, "dheckxKeyExchangeReply exit: state = %d", ss->state);
  return XK_SUCCESS;
}



/* Handle the reply to the key exchange initiation message.
   Called from Demux, with the session (in state DHECKXINITWAIT) and reply message
   as arguments.
   Check that reply message matches our key initiation message.
   Set session state to DHECKXCOMPUTING.
   Computes the session key based on our initiation data and the
   other information in the reply message.
   Enter the computed key in the KM sessions.
   Set session state to DHECKXTRANSPARENT, putting us in transparent mode. */
static xkern_return_t dheckxKeyExchangeDone(sessn,msg) XObj sessn; Msg *msg;
{ SState *ss;
  unsigned long msgbuf[1+1+(2*DEGW)];
  unsigned char key[SAIDMGR_KEYSIZE];
  int t,tmp;

  xTrace2(dheckxspip, TR_MAJOR_EVENTS,
	  "dheckxKeyExchangeDone entered: session = %x, msg = %x", sessn, msg);
  xAssert(xIsSession(sessn));
  ss = sessn->state;
  xTrace2(dheckxspip, TR_MORE_EVENTS, "dheckxKeyExchangeDone: ss = %x, state = %d",
	  ss, ss->state);
  xAssert(ss->state == DHECKXINITWAIT);

  /* Check message length; unpack it. */
  xTrace1(dheckxspip, TR_DETAILED, "dheckxKeyExchangeDone: msgLen = %d", msgLen(msg));
  if ((tmp = msgLen(msg)) != (1+1+2*DEGW)*sizeof(unsigned long))
  { xTrace4(dheckxspip, TR_GROSS_EVENTS,
	"dheckxKeyExchangeDone: Session %x ss %x: Reply message %x length %d is wrong",
	    sessn, ss, msg, tmp);
    return XK_FAILURE; };
  msgPop(msg, dheckxcopy, (char *) msgbuf, (1+1+2*DEGW)*sizeof(unsigned long), 0);

  xIfTrace(dheckxspip, TR_DETAILED)
  {int i; printf("dheckxKeyExchangeInitiate: msg is ");
   for (i=0;i<1+1+(2*DEGW);i++)
     printf("%x:",msgbuf[i]); printf("\n");};

  /* Check msg starts with DHECKXMAGIC, and state of sender is DHECKXREPLYING. */
  if (!(msgbuf[0] == DHECKXMAGIC && msgbuf[1] == DHECKXREPLYING))
  { xTrace5(dheckxspip, TR_GROSS_EVENTS,
	    "dheckxKeyExchangeDone: Session %x ss %x: Reply message %x."
	    " Magic %d or SendState %d is wrong",
	    sessn, ss, msg, msgbuf[0], msgbuf[1]);
    return XK_FAILURE; };

  memcpy(ss->kpx,&msgbuf[2],DEGW * sizeof(unsigned long));
  memcpy(ss->kpy,&msgbuf[2+DEGW],DEGW * sizeof(unsigned long));


 
  ss->state = DHECKXCOMPUTING;  /* Now committed to the Open. */

  Dheckx(ss->kpx,ss->kpy,ss->k,ss->jpx,ss->jpy);

  memcpy(key,ss->jpx,(DEGW-1)*sizeof(int));
  memcpy(&key[(DEGW-1)*sizeof(int)],ss->jpy,(DEGW-1)*sizeof(int));

  dheckxenterkm(sessn, key);

  xIfTrace(dheckxspip, TR_DETAILED)
  { int i; printf("dheckxKeyFinalKey: k*j*p = "); 
  for (i=0;i<DEGW;i++) printf("%x:",ss->jpx[i]);
  for (i=0;i<DEGW;i++) printf("%x:",ss->jpy[i]); printf("\n"); };

  /* set key from gxy */
  xIfTrace(dheckxspip, TR_DETAILED)
  {int i; printf("dheckxKeyExchangeDone: key = ");
   for (i=0;i<8;i++) printf(" %02x", key[i]); printf("\n"); }


  bzero((char *)key,4*sizeof(int)); t=0;  /* Zero out key array on stack. */

  ss->state = DHECKXTRANSPARENT;  /* go transparent */

#ifdef TIME
  xGetTime(&ss->endtime);
  subtime(&ss->starttime,&ss->endtime,&ss->deltatime);
  printf("\nKey exchange completed after %d.%06d seconds.\n",
	 ss->deltatime.sec, ss->deltatime.usec);
#endif

  xTrace1(dheckxspip, TR_EVENTS, "dheckxKeyExchangeDone exit: state = %d", ss->state);
  return XK_SUCCESS;
}


static xkern_return_t romfinddelay(XObj, char **, int, int, void *);

static XObjRomOpt rmopts[1] = {{ "delay", 1, romfinddelay }
			       };

static xkern_return_t
romfinddelay(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  ((PState *)(self->state))->delay = atoi(ptr[2]);
  return XK_SUCCESS;
}


void
dheckxspi_init( self )
     XObj self;
{
    PState	*ps;
    XObj	llp;

    xTrace1(dheckxspip, TR_ERRORS, "dheckx_init entered: protocol = %x", self);
    xAssert(xIsProtocol(self));
    

    if ( ! xIsProtocol(llp = xGetDown(self, TRANSPORT_PROTOCOL)) )
    { xError("dheckx_init: dheckx cannot get lower protocol object"); return; }

    ps = X_NEW(PState);
    self->state = (VOID *)ps;
    ps->actMap = mapCreate(DHECKX_ACTIVE_MAP_SZ, sizeof(ActiveId));
    ps->pasMap = mapCreate(DHECKX_PASSIVE_MAP_SZ, sizeof(PassiveId));
    ps->delay = DHECKX_TIMEOUT;
    ps->flags = 0;


    self->open = dheckxOpen;
    self->openenable = dheckxOpenEnable;
    self->opendisable = dheckxOpenDisable;
    self->opendone = dheckxOpenDone;
    self->close = dheckxClose;  
    self->demux = dheckxDemux;
    findXObjRomOpts(self, rmopts, 1, 0);
    xTrace0(dheckxspip, TR_EVENTS, "dheckx_init exiting");
}

/* Wake up dheckxOpen: Either the session timed out, or the reply arrived.
   evnt is unused, so that Demux can call the same wakeup function. */
static void
dheckxOpenWakeup(evnt, session) Event evnt; void *session;
{ SState	*ss;
  xTrace2(dheckxspip,TR_EVENTS,"dheckxOpenWakeup called; evnt = %x, session = %x", evnt, session);
  xAssert(xIsSession((XObj)session));
  ss = ((XObj)session)->state;
  evCancel(ss->event);  /* cancel timeout event - might be our thread, doesn't matter */
  ss->event = (Event) 0;
  semSignal(&ss->semaphore);
 }

static XObj
dheckxOpen( self, hlpRcv, hlpType, p )
    XObj	self, hlpRcv, hlpType;
    Part 	*p;
{
    XObj 	dheckxSessn;
    XObj	lls;
    PState	*pstate;
    SState	*ss;
    XSaid	lsaid;
    XAlgorithm	algo;
    XIpsecMode	mode;

    xTrace4(dheckxspip,TR_EVENTS,"dheckxOpen entered with args %x %x %x %x",
	    self,hlpRcv,hlpType,p);

    pstate = (PState *) self->state;
    
    if ( p == 0 || partLen(p) < 1 )
    { xTrace0(dheckxspip, TR_SOFT_ERRORS, "dheckxOpen -- bad participants");  
	 return ERR_XOBJ; }


    /* There is no way that there exists allready a active session to a new SAID*/

    lsaid = *((XSaid *)partPop(p[SPI_PART]));
    algo = *((XAlgorithm *)partPop(p[SPI_PART]));
    mode = *((XIpsecMode *)partPop(p[SPI_PART]));

    lls = xOpen(self, self,
		xGetDown(self, TRANSPORT_PROTOCOL), p);
    if ( lls == ERR_XOBJ )
    { xTrace0(dheckxspip, TR_ERRORS, "dheckxOpen: could not open lls"); return ERR_XOBJ; }

    xTrace0(dheckxspip, TR_MAJOR_EVENTS, "creating a new one");
    if ( (dheckxSessn = dheckxCreateSessn(self, hlpRcv, hlpType, lls, lsaid))
		 == ERR_XOBJ )
       { xClose(lls);   return ERR_XOBJ; } 
    
    ss = (SState *)dheckxSessn->state;

    ss->rsaid = 0;
    ss->lsaid = lsaid;
    ss->algorithm = algo;
    ss->mode = mode;

    if (ss->state == DHECKXNEW)
    { ss->state = DHECKXINIT;
      xDuplicate(dheckxSessn); /* Bump the reference count on the session, since
	     it will have message traffic prior to returning from dheckxOpen. */
      dheckxKeyExchangeInitiate(dheckxSessn);
  loop:
      xTrace2(dheckxspip, TR_EVENTS, "dheckxOpen: sleeping for %d seconds; ss = %x",
	      pstate->delay, ss);
      /* schedule time out after DHECKX_TIMEOUT seconds */
      semInit(&ss->semaphore,0);
      ss->event = evSchedule(dheckxOpenWakeup,dheckxSessn,pstate->delay*1000000);
      semWait(&ss->semaphore);  /* the pause that refreshes */
      xTrace2(dheckxspip, TR_EVENTS, "dheckxOpen: awakened; ss = %x, state = %d",
		ss,ss->state);
      /* When we get here, either timed out, or got reply. */      
      if (ss->state == DHECKXCOMPUTING)
      { xTrace1(dheckxspip,TR_EVENTS,
		"dheckxOpen: reply received; still computing key.  ss = %x",ss);
	goto loop; }; /* We are willing to wait forever if computing. */
      if (ss->state == DHECKXINITWAIT)  /* Timed out.  Clean up session & fail. */
      { xTrace1(dheckxspip,TR_ERRORS,"dheckxOpen timed out, returning failure. ss = %x",
		ss);
	dheckxSessn->rcnt--;  dheckxClose(dheckxSessn);  return ERR_XOBJ; };
      if (ss->state == DHECKXNOTACCEPTED)  
	/* Outher side doesn't accept the algorithm. Clean up session & fail. */
      { xTrace1(dheckxspip,TR_ERRORS,"dheckxOpen not accpeted, returning failure. ss = %x",
		ss);
	dheckxSessn->rcnt--;  dheckxClose(dheckxSessn);  return ERR_XOBJ; };
      if (ss->state != DHECKXTRANSPARENT)  /* Bogus state: bomb. */
      { printf("\ndheckxOpen:  Session state %x has bad state value %d\n",
		 ss, ss->state);
	Kabort("Aborting from dheckxOpen."); };
      /* OK!  Drop through. */
      dheckxSessn->rcnt--;  
      /* Drop refcnt to 0; reincremented when I exit dheckxOpen. */
    }

    xTrace1(dheckxspip,TR_EVENTS,"dheckxOpen returning session %x", dheckxSessn);
    return (dheckxSessn);
}


static XObj
dheckxCreateSessn( self, hlpRcv, hlpType, lls, lsaid)
    XObj	self, hlpRcv, hlpType, lls;
    XSaid	lsaid;
{
    XObj	s;
    SState	*ss;
    PState	*ps = (PState *)self->state;
    Part        part[5];  /* So I can partInternalize partBuf, even if 5 participants. */
    char        partBuf[100];

    xTrace5(dheckxspip, TR_EVENTS, "dheckxCreateSessn entered: args %x %x %x %x %u",
	    self, hlpRcv, hlpType, lls, lsaid);

    if (xControl(lls, GETPARTICIPANTS, partBuf, 100) <= 0)
    { xTrace0(dheckxspip, TR_ERRORS, "dheckxCreateSessn: could not get participants");
      return ERR_XOBJ; }
	
    if (partExtLen(partBuf) > 5)
    { xTrace0(dheckxspip, TR_ERRORS,
	      "dheckxCreateSessn: Too many participants;");
      return ERR_XOBJ; }

    partInternalize(part,partBuf);

    if (partLen(part) < 2)
    { xTrace0(dheckxspip, TR_ERRORS, "dheckxCreateSessn: too few participants");
      return ERR_XOBJ; }


    s = xCreateSessn(dheckxSessnInit, hlpRcv, hlpType, self, 1 , &lls);
    s->binding = (Bind) mapBind(ps->actMap,&lsaid , s);

    if ( s->binding == ERR_BIND ) {
	xTrace0(dheckxspip, TR_ERRORS, "dheckxCreateSessn: mapBind error");
	xDestroy(s);
	return ERR_XOBJ;
      }

    ss = X_NEW(SState);
    xTrace1(dheckxspip, TR_FULL_TRACE, "dheckxCreateSessn: X_NEW(SState) returned %x", ss);
    bzero((void *)ss, sizeof(SState));
    ss->state = DHECKXNEW;
    s->state = (VOID *)ss;

    xTrace1(dheckxspip, TR_MORE_EVENTS, "dheckxCreateSessn returning %x",s);
    return s;
}


static xkern_return_t
dheckxOpenEnable(self, hlpRcv, hlpType, part)
    XObj self, hlpRcv, hlpType;
    Part *part;
{
    PState	*ps = (PState *)self->state;
    int 	algo;

    xTrace4(dheckxspip, TR_EVENTS, "dheckxOpenEnable entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, part);

    algo = *((int *)partPop(part[LOCAL_PART]));

    xTrace1(dheckxspip, TR_EVENTS, "dheckxOpenEnable add algo  %u to passiv map",
	    algo);

    mapBind(ps->pasMap,(char *) &algo , hlpRcv);
  

    xTrace0(dheckxspip, TR_MORE_EVENTS, "dheckxOpenEnable returning ");
    return xOpenEnable(self, self, xGetDown(self,TRANSPORT_PROTOCOL),part);
} 


static xkern_return_t
dheckxOpenDisable(self, hlpRcv, hlpType, part)
    XObj self, hlpRcv, hlpType;
    Part *part;
{

    PState	*ps = (PState *)self->state;
    xTrace4(dheckxspip, TR_EVENTS, "dheckxOpenDisable entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, part);

    mapUnbind(ps->pasMap, partPop(part[LOCAL_PART]));

    xTrace0(dheckxspip, TR_MORE_EVENTS, "dheckxOpenDisable returning %x");
    return XK_SUCCESS;
}

static xkern_return_t
dheckxOpenDone(self, lls, llp, hlpType)
    XObj self, lls, llp, hlpType;
{
    return XK_SUCCESS;
}



static xkern_return_t
dheckxDemux( self, lls, msg )
    XObj	self, lls;
    Msg		*msg;
{
  PState	*ps = (PState *)self->state;
  XObj 	s = 0;
  SState	*ss;
  XSaid	lsaid;
  XAlgorithm 	algorithm;
  XIpsecMode	mode;
  xkern_return_t val;
  
  xTrace3(dheckxspip, TR_EVENTS, "dheckxDemux entered: args %x %x %x",
	  self, lls, msg);

  xIfTrace(dheckxspip, TR_DETAILED) msgShow(msg);

  msgPop(msg, dheckxcopy, &lsaid, sizeof(XSaid),NULL);
  msgPop(msg, dheckxcopy, &algorithm, sizeof(XAlgorithm),NULL);
  msgPop(msg, dheckxcopy, &mode, sizeof(XIpsecMode),NULL);
  xDuplicate(lls);

  xTrace3(dheckxspip, TR_EVENTS, "dheckxDemux entered: lsaid algorithm mode %u %u %u ",
	  lsaid, algorithm, mode);

  if ( mapResolve(ps->actMap, &lsaid, &s) == XK_FAILURE ) 
  { 
  	if ( mapResolve(ps->pasMap, &algorithm, &s) == XK_FAILURE ) 
        { 
        Msg 	rmsg;
        XSaid	rsaid;
	xTrace1(dheckxspip, TR_EVENTS, "dheckxDemux: algorithm %u not allowed",algorithm);
	lsaid = 0;
  	msgPop(msg, dheckxcopy, &rsaid, sizeof(XSaid),NULL);
        msgConstructEmpty(&rmsg);
  	msgPush(&rmsg, (MStoreFun) dheckxstore, &lsaid, sizeof(XSaid), NULL);
  	msgPush(&rmsg, (MStoreFun) dheckxstore, &algorithm, sizeof(XSaid), NULL);
  	msgPush(&rmsg, (MStoreFun) dheckxstore, &mode, sizeof(XIpsecMode), NULL);
  	msgPush(&rmsg, (MStoreFun) dheckxstore, &rsaid, sizeof(XSaid), NULL);
	/* sending return message ?? */
  	return XK_SUCCESS; 
	}
	else
	{

		lsaid = (int) xOpenDone(s,self,xGetDown(self,0)); /*hack for lls*/
    		if ( (s = dheckxCreateSessn(self, s, s, lls, lsaid)) == ERR_XOBJ )
      		 { xClose(lls);   return XK_FAILURE; } 
  		ss = (SState *)s->state;
		if (ss->state == DHECKXNEW)
		{
			ss->state = DHECKXRESPONDER;
		}
		else
		{
			return XK_FAILURE;
		}
    		ss->lsaid = lsaid;
   		ss->algorithm =  algorithm;
   		ss->mode =  mode;
	}
  }
  else
  	ss = (SState *)s->state;

  msgPop(msg, dheckxcopy, &ss->rsaid, sizeof(XSaid),NULL);

  xTrace1(dheckxspip, TR_EVENTS, "dheckxDemux entered: rsaid %u ",
	  ss->rsaid);

  if ( xIsSession(s) )  val = xPop(s, lls, msg, 0);
  else val = XK_FAILURE;

  return val;
}


static xkern_return_t
dheckxClose( s )
    XObj	s;
{
    PState	*ps = (PState *)xMyProtl(s)->state;
    xkern_return_t res;
    SState	*ss = (SState *)s->state;

    xTrace1(dheckxspip, TR_MAJOR_EVENTS, "dheckxClose: closing session %x", s);
    xAssert( xIsSession( s ) );
    if (s->rcnt == 0)
    {
    	res = mapUnbind( ps->actMap,&ss->lsaid);
    	xAssert( res != XK_FAILURE );
    	/* Clear out SState. */
    	if (ss = s->state)
    	{
      	/* I don't have to do anything about semaphore & event, because they
	 	are only used while waiting for reply & computing, and nobody will
		 have the session pointer to call Close with. */
      		bzero((void *)ss,sizeof(SState)); }   /* The xDestroy will xFree(ss). */
    	xClose(xGetDown(s,TRANSPORT_PROTOCOL));
    	xDestroy( s );
    }
    xTrace0(dheckxspip, TR_DETAILED, "dheckxClose returning");
    return XK_SUCCESS;
}

static
long dheckxcopy(to,from,len,ignore)
    void *to; char *from; void *ignore; long len;
{
  xTrace4(dheckxspip, TR_EVENTS, "dheckxcopy args: %x %x %x %x", to,from,len,ignore);

  bcopy(from,to,len);  return len;
}

static
long dheckxstore(from,to,len,ignore)
    void *to; char *from; void *ignore; long len;
{
  xTrace4(dheckxspip, TR_EVENTS, "dheckxstore args: %x %x %x %x", to,from,len,ignore);

  bcopy(from,to,len);  return len;
}




static xkern_return_t
dheckxPop( s, lls, m, hdr)
    XObj	s, lls;
    Msg 	*m;
    void        *hdr;
{ SState *ss = (SState *)s->state;

  xTrace5(dheckxspip, TR_EVENTS, "dheckxPop: args %x %x %x %x.  state %d.",
	  s, lls, m, hdr, ss->state);
  switch (ss->state)
  {
    case DHECKXINITWAIT:     /* process the reply to our offer message */
      if (ss->rsaid == 0)   /* offer was rejected */
      {
	 ss->state = DHECKXNOTACCEPTED;
	 dheckxOpenWakeup(0,s);
	 return XK_FAILURE;
      }
      if (dheckxKeyExchangeDone(s, m) == XK_SUCCESS)
	{ dheckxOpenWakeup(0,s); return XK_SUCCESS;} /* if ok, wakeup Open */
      else return XK_FAILURE;
      /* Fail if something wrong with reply; Open will eventually time out anyway. */

    case DHECKXRESPONDER:    /* reply to an offer message */
      dheckxKeyExchangeReply(s, m);
      return XK_SUCCESS; /* close session after key exchange */

    /* Other possible states:
       REPLYING, if the replier is slow computing the key.
       COMPUTING, possibly. */

    default:
      xTrace2(dheckxspip, TR_ERRORS,
	      "dheckxPop: Unexpected session state %x, dropping message %x",
	      ss->state, m);
      return XK_FAILURE;  }
}

static void
dheckxSessnInit(s)
    XObj s;
{ xTrace1(dheckxspip, TR_EVENTS, "dheckxSessnInit: session %x", s);
  s->close = dheckxClose;
  s->pop = dheckxPop;
}

