/*
 * $RCSfile: ipcp_config.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1994  Arizona Board of Regents
 *
 * $Log: ipcp_config.c,v $
 * Revision 1.3  1994/10/28 20:35:28  andrey
 * *** empty log message ***
 *
 * Revision 1.2  1994/10/27  17:26:00  andrey
 * removed non-ansi definitions
 *
 * Revision 1.1  1994/10/24  20:48:44  andrey
 * Initial revision
 *
 *
 * $Revision: 1.3 $
 * $Date: 1994/10/28 20:35:28 $
 */


/*
 * IPCP negotiation processing routines
 */

/*
 * see if peer's parameters are ok
 * return anything unpalatable
 * need to check if we've rejected any of these previously
 */
static xkern_return_t
options_acceptable(
  XObj sess,
  PPP_options * lo,
  PPP_options ** rej_opts,
  PPP_options ** nak_opts,
  int rej_not_nak)
{
  opt_list_elem * ole = 0;
  *rej_opts = 0;
  *nak_opts = 0;

  ole = first_option(lo);
  while (ole) {
    switch(ole->type) {

      case IPCP_ADDRESS:
      break;
      case IPCP_COMPRESSION:
      break;

      case IPCP_UNKNOWN:		/* reject it straight away */
        if (!*rej_opts) *rej_opts = X_NEW(PPP_options);
        ppp_add_option(*rej_opts,
                   ipcp_duplicate_option(ole->opt_ptr, ole->type),
                   ole->type);
      break;
    }
    ole = next_option(ole);
  }
  if (*rej_opts || *nak_opts) {
    if (*rej_opts && *nak_opts) {
       ppp_destroy_options(*nak_opts, ipcp_destroy_opt);
       nak_opts = 0;
    }
    return XK_FAILURE;
  }
  return XK_SUCCESS;
}

/*
 * recalculate offered nak'd options
 */
static PPP_options *
IPCP_options_reconsider(XObj sess, PPP_options * lo)
{
  PPP_options * retval = X_NEW(PPP_options);
  opt_list_elem * ole = first_option(lo);

  while (ole) {
    switch(ole->type) {
      case IPCP_ADDRESS:
      break;
      case IPCP_COMPRESSION:
      break;
    }
    ole = next_option(ole);
  }
  return retval;
}

static void
IPCP_opt_store( void * src, char * dest, long count, void * type)
{
  int ipcp_t = (int) type;

  switch(ipcp_t) {
    case IPCP_ADDRESS:
      bcopy(src, (char *)dest, sizeof(IPhost));
    break;

    case IPCP_COMPRESSION:
      bcopy(src, (char *)dest, sizeof(COMPRESSION_opt));
    break;
  }
}

static long
IPCP_opt_load( void * dest, char * src, long count, void * type)
{
  int ipcp_t = (int) type;
  int copied = -1;

  switch(ipcp_t) {
    case IPCP_ADDRESS:
      bcopy(src, (char *)dest, sizeof(IPhost));
      copied = sizeof(IPhost);
    break;

    case IPCP_COMPRESSION:
      bcopy(src, (char *)dest, sizeof(COMPRESSION_opt));
      copied = sizeof(COMPRESSION_opt);
    break;

  }
  return copied;
}

/*
 * turns an PPP_options list object into a Msg
 */
static Msg *
make_config_msg(PPP_options * this)
{
  Msg * new_msg = X_NEW(Msg);
  PPP_opt lo;
  opt_list_elem * ole;
  int size = 0;

  msgConstructEmpty(new_msg);
  ole = first_option(this);

  while (ole) {
    switch(ole->type) {
      case IPCP_ADDRESS:
        size = sizeof(ADDRESS_opt);
        ADDRESS_opt_init(&lo);
      break;

      case IPCP_COMPRESSION:
        size = sizeof(COMPRESSION_opt);
        COMPRESSION_opt_init(&lo);
      break;

      case IPCP_UNKNOWN:
        {
          UNKNOWN_opt * uop = (UNKNOWN_opt *)ole->opt_ptr;
          xTrace1(ipcpp, TR_SOFT_ERRORS,
                "make_msg: odd configuration option %d", uop->type);
          lo.type = uop->type;
          lo.length = uop->length;
          msgAppend(new_msg, PPP_opt_hdr_store, &lo, sizeof(PPP_opt), 0, 0);
          msgAppend(new_msg, IPCP_opt_store, uop->buf, lo.length,
                    (void*)IPCP_UNKNOWN, 0);
          goto skip;
        }
    }
    msgAppend(new_msg, PPP_opt_hdr_store, &lo, sizeof(PPP_opt), 0, 0);
    if (size) {
      msgAppend(new_msg, IPCP_opt_store, ole->opt_ptr, size,
                (void *)(int)ole->type, 1);
    }

skip:
    ole = next_option(ole);
  }
  return new_msg;
}

/*
 * turns a Msg object into a PPP_options object, preserving order
 */
static PPP_options *
make_config_options( Msg * this, int len)
{
  PPP_options * new_options = X_NEW(PPP_options);
  PPP_options_init(new_options);

  while(len > 0) {
    PPP_opt lo;
    void * o;

    /* return null if we can't tear down packet? */
    if (msgPop(this, PPP_opt_hdr_load, &lo, sizeof(PPP_opt), 0) == FALSE) {
       ppp_destroy_options(new_options, ipcp_destroy_opt);
       return 0;
    } 
    len -= sizeof(PPP_opt);

    switch(lo.type) {
    case IPCP_ADDRESS:
      o = X_NEW(ADDRESS_opt);
      if (msgPop(this, IPCP_opt_load, o, sizeof(ADDRESS_opt),
                 (void *) IPCP_ADDRESS) == FALSE) {
        ppp_destroy_options(new_options, ipcp_destroy_opt);
        return 0;
      }
      len -= sizeof(ADDRESS_opt);
    break;

    case IPCP_COMPRESSION:
      o = X_NEW(COMPRESSION_opt);
      if (msgPop(this, IPCP_opt_load, o, sizeof(COMPRESSION_opt),
                 (void *)IPCP_COMPRESSION) == FALSE) {
        ppp_destroy_options(new_options, ipcp_destroy_opt);
        return 0;
      }
      len -= sizeof(COMPRESSION_opt);
    break;

      default:
        {
          UNKNOWN_opt * uop;
          xTrace1(ipcpp, TR_SOFT_ERRORS,
                "make_config_options: odd configuration option %d", lo.type);
          uop = o = X_NEW(UNKNOWN_opt);
          uop->type = lo.type;
          uop->length = lo.length;
          uop->buf = xMalloc(lo.length);
          msgPop(this, IPCP_opt_load, uop->buf, lo.length,
                 (void *)IPCP_UNKNOWN);
          ppp_add_option(new_options, o, IPCP_UNKNOWN);
          len -= lo.length;
          goto skip;
        }
    }
    ppp_add_option(new_options, o, lo.type);

skip:
    ;
  }
  return new_options;
}
