/*     
 * $RCSfile: lcp.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1994  Arizona Board of Regents
 *
 * $Log: lcp.c,v $
 * Revision 1.2  1994/10/27 17:27:04  andrey
 * removed non-ansi definitions
 *
 * Revision 1.1  1994/10/24  20:46:02  andrey
 * Initial revision
 *
 *
 * $Revision: 1.2 $
 * $Date: 1994/10/27 17:27:04 $
 */

/*
 * link control protocol for use with PPP
 */

int     tracelcpp;

#include "xkernel.h"
#include "ppp.h"
#include "lcp.h"
#include "lcp_constants.h"

#include	"../ppp/ppp_options.h"
#include	"../ppp/ppp_fsm.h"
#include	"lcp_i.h"

static void             getproc_protl(XObj);
static void             getproc_sessn(XObj);
extern xkern_return_t   lcpProcessOpts(XObj);
static xkern_return_t   lcpCloseProtl( XObj );
static xkern_return_t   lcpCloseSessn( XObj );
static int              lcpControlProtl( XObj, int, char *, int ) ;
static int              lcpControlSessn( XObj, int, char *, int ) ;
static XObj             lcpCreateSessn( XObj, XObj, XObj, XObj );
static xkern_return_t   lcpDemux( XObj, XObj, Msg * );
static XObj             lcpOpen( XObj, XObj, XObj, Part * );
static xkern_return_t   lcpOpenDisable( XObj, XObj, XObj, Part * );
static xkern_return_t   lcpOpenEnable( XObj, XObj, XObj, Part * );
static xkern_return_t   lcpOpenDisableAll( XObj, XObj );

static xkern_return_t   lcpOpenDone( XObj, XObj, XObj, XObj);
static xkern_return_t   lcpCloseDone( XObj );

static xkern_return_t   lcpPop( XObj, XObj, Msg *, void * );
/* static xmsg_handle_t    lcpPush( XObj, Msg * );	*/
static xkern_return_t	process_message(XObj, Msg *);
static void		SSTATE_init(SSTATE *, XObj);

#include        "lcp_fsm_table.h"

#include	"lcp_options.c"
#include	"lcp_config.c"

void
lcp_init(XObj self)
{
  PSTATE * pstate = X_NEW(PSTATE);
  XObj     llp;

  PSTATE_init(pstate);
  xTraceP0(self, TR_GROSS_EVENTS, "lcp_init");

  getproc_protl(self);
  self->state = (void *)pstate;
  (void)lcpProcessOpts(self);
  xIfTraceP(self, TR_DETAILED) {
    ppp_print_options(&pstate->rom_lcp_options, lcp_print_opt);
    printf("\n");
  }

  /* should ensure that lower protocol is actually PPP */

  if ( ! xIsProtocol(llp = xGetDown(self, 0)) ) {
    Kabort("LCP could not get lower protocol");
  }

  if ( xOpenEnable(self, self, xGetDown(self, 0), 0) == XK_FAILURE) {
    xTraceP0(self, TR_ALWAYS, "init: can't openenable lower protocol");
    FREE(pstate);
    self->state = 0;
    return;
  }
  xTraceP0(self, TR_EVENTS, "init done");
}

static void
getproc_protl(XObj s)
{
  xAssert(xIsProtocol(s));
  s->close = lcpCloseProtl;
  s->control = lcpControlProtl;
  s->open = lcpOpen;
  s->opendone = lcpOpenDone;
  s->openenable = lcpOpenEnable;
  s->opendisable = lcpOpenDisable;
  s->closedone = lcpCloseDone;
  s->demux = lcpDemux;
  s->opendisableall = lcpOpenDisableAll;
}

static void
getproc_sessn(XObj s)
{
  xAssert(xIsSession(s));
  s->pop = lcpPop;
  s->control = lcpControlSessn;
  s->close = lcpCloseSessn;
}

static xkern_return_t
lcpCloseProtl(XObj self)
{
  xTraceP0(self, TR_EVENTS, "CloseProtl");
  return XK_FAILURE;
}

/*
 * should be via lower protocol
 */
static xkern_return_t
lcpCloseSessn(XObj self)
{
  xTraceP0(self, TR_EVENTS, "CloseSessn");
  return XK_FAILURE;
}

static int
lcpControlProtl(XObj self, int opcode, char * buf, int len)
{
  PSTATE * pstate = (PSTATE *)self->state;
  xTraceP0(self, TR_EVENTS, "ControlProtl");

  switch(opcode) {

    case PPP_NEGOTIATE:
      return lcpControlSessn(pstate->TheSession, opcode, buf, len);
    break;

  }
  return 0;
}

static int
lcpControlSessn(XObj self, int opcode, char * buf, int len)
{
  SSTATE * sstate = (SSTATE *) self->state;
  FSM_info * fi = sstate->lcp_fsm;

  xTraceP0(self, TR_EVENTS, "ControlSessn");

  switch(opcode) {

    case PPP_NEGOTIATE:
      if (len < sizeof(sstate->my_agreed_options))
        return 0;
      ppp_process_event(fi, F_open);
      fi->waiting_on_nego = 1;
      semWait(&fi->Nego_sem);
      if (fi->nego_ok) {
        bcopy(buf, (char *)sstate->my_agreed_options,
                  sizeof(sstate->my_agreed_options));
        return sizeof(sstate->my_agreed_options);
      }
      else
        return 0;
    break;

  }
  return 0;
}

/*
 * always the result of an OpenDone
 */
static XObj
lcpCreateSessn(XObj self, XObj hlpRcv, XObj hlpType, XObj lls)
{
  XObj s;
  SSTATE * sstate = (SSTATE *)X_NEW(SSTATE);

  xTraceP0(self, TR_EVENTS, "CreateSessn");
  s = xCreateSessn(getproc_sessn, hlpRcv, hlpType, self, 1, &lls);
  s->state = sstate;
  SSTATE_init(sstate, s);

  sstate->my_nego_options = X_NEW(PPP_options);
  PPP_options_init(sstate->my_nego_options);
  ppp_copy_options(sstate->my_nego_options,
           &((PSTATE *)self->state)->rom_lcp_options,
           lcp_duplicate_option);

  xIfTraceP(s, TR_DETAILED) {
    printf("session options\n");
    ppp_print_options(sstate->my_nego_options, lcp_print_opt);
    printf("\n");
  }
  return s;
}

/*
 * Corresponds to a DOWN event
 */
static xkern_return_t
lcpCloseDone(XObj self)
{
  FSM_info * fi = ((SSTATE *)self->state)->lcp_fsm;

  xTraceP0(self, TR_EVENTS, "CloseDone");
  xSetDown(self, 0, ERR_XOBJ);
  ppp_process_event(fi, F_down);
  return XK_SUCCESS;
}

/*
 * Corresponds to an UP event
 */
static xkern_return_t
lcpOpenDone(XObj self, XObj lls, XObj llp, XObj hlpType)
{
  PSTATE * pstate = (PSTATE *)self->state;
  FSM_info * fi;

  xTraceP0(self, TR_EVENTS, "OpenDone");
  if (pstate->TheSession == ERR_XOBJ) {
    pstate->TheSession = lcpCreateSessn(self, self, self, lls);
    xDuplicate(pstate->TheSession);
  }
  else {
    xSetDown(pstate->TheSession, 0, lls);
  }
  fi = ((SSTATE*)pstate->TheSession->state)->lcp_fsm;
  ppp_process_event(fi, F_up);
  return XK_SUCCESS;
}

static xkern_return_t
lcpDemux(XObj self, XObj lls, Msg * msg)
{
  PSTATE * pstate = (PSTATE *)self->state;
  FSM_info * fi = ((SSTATE*)pstate->TheSession->state)->lcp_fsm;

  xTraceP0(self, TR_EVENTS, "Demux");

  switch (fi->state) {
    case F_initial:
    break;
    case F_starting:
    break;
    case F_closed:
      ppp_process_event(fi, F_open);
    break;
    case F_stopped:
    break;
    case F_closing:
    break;
    case F_stopping:
    break;

    default:
    break;
  }
  return xPop(pstate->TheSession, lls, msg, 0);
}

static XObj
lcpOpen(XObj self, XObj hlpRcv, XObj hlpType, Part * p)
{
  xTraceP0(self, TR_EVENTS, "Open");
  return ERR_XOBJ;
}

static xkern_return_t
lcpOpenDisable(XObj self, XObj hlpRcv, XObj hlpType, Part * p)
{
  xTraceP0(self, TR_EVENTS, "OpenDisable");
  return XK_FAILURE;
}

static xkern_return_t
lcpOpenEnable(XObj self, XObj hlpRcv, XObj hlpType, Part * p)
{
  xTraceP0(self, TR_EVENTS, "OpenEnable");
  return XK_FAILURE;
}

static xkern_return_t
lcpOpenDisableAll(XObj self, XObj hlpRcv)
{
  xTraceP0(self, TR_EVENTS, "OpenDisableAll");
  return XK_FAILURE;
}

static xkern_return_t
lcpPop( XObj self, XObj ds, Msg * msg, void * hdr)
{
  xTraceP0(self, TR_EVENTS, "Pop");

  /* had better be no arg, and no ds */
  return process_message(self, msg);
}

/*
 * Push creates an echo request or a discard packet 
static xmsg_handle_t
lcpPush(sess, msg)
 */

/* ------------------------------------------------------------------------ */
/*                          FSM routines                                    */
/* ------------------------------------------------------------------------ */

static void
signal_negotiation_done(XObj sess)
{
  SSTATE * sstate = (SSTATE *) sess->state;
  FSM_info * fi = sstate->lcp_fsm;

  if (fi->nego_ok)
    xControl(xGetDown(sess, 0), PPP_DONE_NEGOTIATION,
           (void *)&sstate->my_agreed_options,
            sizeof(&sstate->my_agreed_options));
  else
    xControl(xGetDown(sess, 0), PPP_DONE_NEGOTIATION, (void *)0, 0);
}

static Msg *
make_config_req_msg(XObj sess)
{
  SSTATE * sstate = (SSTATE *)sess->state;

  return  make_config_msg(sstate->my_nego_options);
}

static Msg *
make_config_ack_msg(XObj sess)
{
  SSTATE * sstate = (SSTATE *)sess->state;

  return  make_config_msg(sstate->peer_agreed_options);
}

static Msg *
make_config_nak_msg(XObj sess)
{
  SSTATE * sstate = (SSTATE *)sess->state;

  return  make_config_msg(sstate->peer_nego_options);
}

static Msg *
make_config_rej_msg(XObj sess)
{
  SSTATE * sstate = (SSTATE *)sess->state;

  return  make_config_msg(sstate->peer_un_nego_options);
}

/*
 * pull apart message, then turn crank on automaton
 */
static xkern_return_t
process_message(XObj sess, Msg * msg)
{
  int code, len, id;
  fsm_event next_event = NO_EVENT;
  SSTATE * sstate = (SSTATE *)sess->state;
  FSM_info * fi = sstate->lcp_fsm;
  Msg * msg_copy = X_NEW(Msg);

  /* destroyed after action is performed above */
  msgConstructCopy(msg_copy, msg);

  code = ppp_decode_packet(msg, &id, &len);

  xTraceP3(sess, TR_DETAILED, "process message type %x id %x len %d",
          code, id, len);

  if (msgLen(msg) + sizeof(PPP_header) != len) {
    xTraceP2(sess, TR_SOFT_ERRORS, "length %d at odds with msg length %d",
            len, msgLen(msg) + sizeof(PPP_header) );
/*
 * just discard
    return;
 */
  }
  len -= sizeof(PPP_header);

  switch(code) {

    /* generate either F_rcv_cfg_req_good or F_rcv_cfg_req_bad */
    case PPP_CONF_REQ:
      {
        PPP_options * lo;

        lo = make_config_options(msg, len);

        xIfTraceP(sess, TR_DETAILED) {
          printf("Received options:\n");
          ppp_print_options(lo, lcp_print_opt);
          printf("\n");
        }
        if (options_acceptable(sess, lo, 
                               &sstate->peer_un_nego_options,
                               &sstate->peer_nego_options, 1) == XK_FAILURE) {
          if (sstate->peer_un_nego_options) {
            xIfTraceP(sess, TR_DETAILED) {
              printf("I rejected options:\n");
              ppp_print_options(sstate->peer_un_nego_options, lcp_print_opt);
              printf("\n");
            }
            fi->nak_not_rej = 0;
          }
          else {
            xIfTraceP(sess, TR_DETAILED) {
              printf("I nak'ed options:\n");
              ppp_print_options(sstate->peer_nego_options, lcp_print_opt);
              printf("\n");
            }
            fi->nak_not_rej = 1;
          }
          next_event = F_rcv_cfg_req_bad;
        }
        else {
          xTraceP0(sess, TR_DETAILED, "options ok");
          sstate->peer_agreed_options = lo;
          next_event = F_rcv_cfg_req_good;
        }
      }
    break;

    case PPP_CONF_ACK:		/* generate F_rcv_cfg_ack */
      {
        PPP_options * lo = make_config_options(msg, len);
        xIfTraceP(sess, TR_DETAILED) {
          printf("peer ack'ed options:\n");
          ppp_print_options(lo, lcp_print_opt);
          printf("\n");
        }

        if (sstate->my_nego_options->ident != id) {
          xTraceP2(sess, TR_SOFT_ERRORS,
                  "questionable nack/rej - sent id %d, received %d",
                  sstate->my_nego_options->ident, id);
          /* ignore packet
             return;
*/
        }
        if (!ppp_options_match(sstate->my_nego_options,
                               lo, lcp_option_matches)) {
          xTraceP0(sess, TR_SOFT_ERRORS, "questionable ack - option mismatch");
          /* ignore packet
             return;
*/
        }
        if (sstate->my_agreed_options == 0)
          sstate->my_agreed_options = lo;
        else {
          ppp_options_append(sstate->my_agreed_options, lo);
        }
        next_event = F_rcv_cfg_ack;
      }
    break;

    case PPP_CONF_REJ:		/* generate F_rcv_cfg_nak */
      {

        PPP_options * new_opts, * lo;
        if (sstate->my_nego_options->ident != id) {
          xTraceP2(sess, TR_SOFT_ERRORS,
                  "questionable rej - sent id %d, received %d",
                  sstate->my_nego_options->ident, id);
          /* ignore packet
             return;
*/
        }
        lo = make_config_options(msg, len);
        xIfTraceP(sess, TR_DETAILED) {
          printf("peer rejected options:\n");
          ppp_print_options(lo, lcp_print_opt);
          printf("\n");
        }

        /* retire rejected options */
        new_opts = ppp_option_differences(sstate->my_nego_options, lo,
                                          lcp_option_matches,
                                          lcp_duplicate_option);
        /* no options remain? */
        if (!new_opts) {
          new_opts = X_NEW(PPP_options);
        }
        new_opts->ident = id++;
        sstate->my_un_nego_options = lo;
        ppp_destroy_options(sstate->my_nego_options, lcp_destroy_opt);

        xIfTraceP(sess, TR_DETAILED) {
          printf("sending options:\n");
          ppp_print_options(new_opts, lcp_print_opt);
          printf("\n");
        }
        sstate->my_nego_options = new_opts;
        next_event = F_rcv_cfg_nak;
      }
    break;

    case PPP_CONF_NAK:		/* generate F_rcv_cfg_nak */
      {
        PPP_options * new_opts, * lo;
        PPP_options * diff;
        if (sstate->my_nego_options->ident != id) {
          xTraceP2(sess, TR_SOFT_ERRORS,
                  "questionable nack/rej - sent id %d, received %d",
                  sstate->my_nego_options->ident, id);
          /* ignore packet
             return;
*/
        }
        /* recalculate options */
        lo = make_config_options(msg, len);
        xIfTraceP(sess, TR_DETAILED) {
          printf("peer nak'ed options:\n");
          ppp_print_options(lo, lcp_print_opt);
          printf("\n");
        }

        new_opts = LCP_options_reconsider(sess, lo);
        diff = ppp_option_differences(sstate->my_nego_options, new_opts,
                                      lcp_option_matches,
                                      lcp_duplicate_option);
        ppp_options_append(new_opts, diff);
        xIfTraceP(sess, TR_DETAILED) {
          printf("sending options:\n");
          ppp_print_options(new_opts, lcp_print_opt);
          printf("\n");
        }
        new_opts->ident = id++;
        ppp_destroy_options(sstate->my_nego_options, lcp_destroy_opt);
        sstate->my_nego_options = new_opts;
        next_event = F_rcv_cfg_nak;
      }
    break;

    case PPP_TERM_REQ:		/* generate F_rcv_term_req */
      next_event = F_rcv_term_req;
    break;

    case PPP_TERM_ACK:		/* generate F_rcv_term_ack */
      next_event = F_rcv_term_ack;
    break;

    /* generate LCP_rcv_cod_rej_perm (RXJ+) */
    /*       or LCP_rcv_cod_rej_catas (RXJ-) */
    case LCP_CODE_REJ:
    case LCP_PROT_REJ:

    case LCP_ECHO_REQ:		/* generate LCP_rcv_echo_req */
      next_event = F_rcv_echo_req;
    break;

    case LCP_ECHO_REP:		/* drop packet on floor */
    case LCP_DISC_REQ:
    break;

    default:			/* generate LCP_rcv_unk_code */
      next_event = F_rcv_unk_code;
      /* dump the code field */
      msgPopDiscard(msg_copy, 1);
      fi->msg = msg_copy;
    break;
  }
  ppp_process_event(fi, next_event);
  return XK_SUCCESS;
}


/* ------------------------------------------------------------------------ */

static void
FSM_info_init(
  FSM_info * fi,
  XObj sess,
  xkern_return_t (*pm)(XObj, Msg *),
  Msg * (*mcrm)(XObj),
  Msg * (*mcam)(XObj),
  Msg * (*mcnm)(XObj),
  Msg * (*mcjm)(XObj),
  void (*snd)(XObj))
{
  fi->session = sess;
  fi->state_table = &lcp_state_table;
  fi->action_table = &lcp_fsm_action;
  fi->timerev = 0;
  semInit(&fi->Nego_sem, 0);
  fi->state = F_initial;
  fi->msg = 0;
  fi->restart_counter = 0;
  fi->timerev = 0;
  fi->waiting_on_nego = 0;
  fi->process_message = pm;
  fi->make_config_req_msg = mcrm;
  fi->make_config_ack_msg = mcam;
  fi->make_config_nak_msg = mcnm;
  fi->make_config_rej_msg = mcjm;
  fi->signal_negotiation_done = snd;
}

static void
SSTATE_init(SSTATE * ssp, XObj sess)
{
  ssp->lcp_fsm = X_NEW(FSM_info);
  FSM_info_init(ssp->lcp_fsm, sess,
      process_message, make_config_req_msg,
      make_config_ack_msg, make_config_nak_msg,
      make_config_rej_msg, signal_negotiation_done);

  ssp->lcp_fsm->time_options = &((PSTATE*)sess->myprotl->state)->implem_options;
  ssp->code_reject_id = 133;		/* make it unusual while testing */
  ssp->magic_number = 73733;

  ssp->my_agreed_options = 0;
  ssp->my_nego_options = 0;
  ssp->my_un_nego_options = 0;
  ssp->peer_agreed_options = 0;
  ssp->peer_nego_options = 0;
  ssp->peer_un_nego_options = 0;
}

