/*
 * $RCSfile: ppp_options.h,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1994  Arizona Board of Regents
 *
 * $Log: ppp_options.h,v $
 * Revision 1.2  1994/10/27 17:27:28  andrey
 * removed non-ansi definitions
 *
 * Revision 1.1  1994/10/24  20:44:34  andrey
 * Initial revision
 *
 *
 * $Revision: 1.2 $
 * $Date: 1994/10/27 17:27:28 $
 */

/*
 * common code for protocol control protocol negotiation
 */

#define PPP_CONF_REQ    0x1    /* Configure Request                     */
#define PPP_CONF_ACK    0x2    /* Configure-Ack                         */
#define PPP_CONF_NAK    0x3    /* Configure-Nak                         */
#define PPP_CONF_REJ    0x4    /* Configure Reject                      */
#define PPP_TERM_REQ    0x5    /* Terminate Request                     */
#define PPP_TERM_ACK    0x6    /* Terminate Ack                         */

typedef struct
{
  u_char         code;
  u_char         ident;
  u_short        length;
} PPP_header;

#define PPP_CONF_REQ_init(PPP_header_ptr) ((PPP_header_ptr)->code=PPP_CONF_REQ)
#define PPP_CONF_ACK_init(PPP_header_ptr) ((PPP_header_ptr)->code=PPP_CONF_ACK)
#define PPP_CONF_NAK_init(PPP_header_ptr) ((PPP_header_ptr)->code=PPP_CONF_NAK)
#define PPP_CONF_REJ_init(PPP_header_ptr) ((PPP_header_ptr)->code=PPP_CONF_REJ)

#define PPP_TERM_REQ_init(PPP_header_ptr) \
    ((PPP_header_ptr)->code=PPP_TERM_REQ, (PPP_header_ptr)->length=4)

#define PPP_TERM_ACK_init(PPP_header_ptr)\
    ((PPP_header_ptr)->code=PPP_TERM_ACK, (PPP_header_ptr)->length=4)


/* -------------------------------------------------------------------------*/
/*                     option header stuff                                  */
/* -------------------------------------------------------------------------*/


typedef struct
{
  u_char type;
  u_char length;
} PPP_opt;

typedef struct
{
  u_char         length;
  u_char         type;                   /* number of unclassifiable option */
  char *        buf;
} UNKNOWN_opt;

#define free_unknown(UNKOWN_opt_ptr)\
  (FREE((UNKOWN_opt_ptr)->buf), FREE(UNKOWN_opt_ptr))

#define PPP_options_init(options_ptr) \
  ((options_ptr)->ident=0, (options_ptr)->first_opt=0)

long PPP_opt_hdr_load(void * dest, char * src, long count, void * type);

void PPP_opt_hdr_store(void * src, char * dest, long count, void * type);

long PPP_hdr_load(void * dest, char * src, long count, void * type);

void PPP_hdr_store(void * src, char * dest, long count, void * type);

/* -------------------------------------------------------------------------*/
/*                     options handling stuff                               */
/* -------------------------------------------------------------------------*/

typedef struct opt_list_elem {
  void * opt_ptr;
  u_char type;
  struct opt_list_elem * next;
  struct opt_list_elem * prev;
} opt_list_elem;

#define opt_list_elem_init(ole_ptr)\
  ((ole_ptr)->next = (ole_ptr)->prev = 0,(ole_ptr)->opt_ptr=0,(ole_ptr)->type=0)

#define link_ole(ole_ptr1, ole_ptr2)\
  ((ole_ptr2)->next = (ole_ptr1)->next,\
   (ole_ptr2)->prev = (ole_ptr1),\
   (ole_ptr1)->next = (ole_ptr2))

typedef struct {
  u_char ident;          /* gotta keep around for ack, naks, etc */
  opt_list_elem * first_opt;
} PPP_options;

#define first_option(options_ptr) ((options_ptr)->first_opt)
#define next_option(ole_ptr) ((ole_ptr)->next)

#ifdef	__STDC__

void ppp_add_option(PPP_options * , void *, u_char);

void ppp_copy_options(PPP_options * to, PPP_options * from,
                      void * (*duplicate_option)(void * , u_char ));

void ppp_destroy_options(PPP_options * opt,
                         void (*destroy_opt)(void * , u_char ));

void ppp_options_append( PPP_options * to, PPP_options * from );

int ppp_options_match( PPP_options * , PPP_options * ,
                       int (*option_matches)(void * , void * , u_char ));

PPP_options * ppp_option_differences(PPP_options * , PPP_options * ,
                       int (*option_matches)(void * , void * , u_char ), 
                       void * (*duplicate_option)(void * , u_char ));

void ppp_print_options(PPP_options * opt, void (*print_opt)(void *, u_char));

#else

void ppp_add_option( );

void ppp_copy_options( );

void ppp_destroy_options( );

void ppp_options_append( );

int ppp_options_match( );

PPP_options * ppp_option_differences( );

void ppp_print_options( );

#endif /* __STDC__ */
