/*
 * $RCSfile: ppp_rom.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1994  Arizona Board of Regents
 *
 * $Log: ppp_rom.c,v $
 * Revision 1.2  1994/10/27 17:27:28  andrey
 * removed non-ansi definitions
 *
 * Revision 1.1  1994/10/24  20:44:34  andrey
 * Initial revision
 *
 *
 * $Revision: 1.2 $
 * $Date: 1994/10/27 17:27:28 $
 */

/*
 * ppp options processing
 *   fixed association of an ip address with ppp protocol
 *   so that ppp can look like ARP as well as ETH
 */

#include "xkernel.h"
#include "romopt.h"
#include "ppp_internal.h"
#include "ppp_constants.h"

static xkern_return_t   ppp_rom_opts( XObj, char **, int, int, VOID * );

extern int tracepppp;

/*
 * ppp addr LocalIPaddr RemoteIPaddr devicename mtu
 * ppp addr 192.12.69.119 192.12.69.124 /dev/ttya 1000
* or
 * ppp addr LocalIPaddr RemoteIPaddr udpportnum mtu
 * ppp addr 192.12.69.119 192.12.69.124 9999 1000
 */

static XObjRomOpt	rOpts[] =
{
  {  "ipaddr", 6, ppp_rom_opts }
};

xkern_return_t
pppProcessOpts(self)
  XObj self;
{
  return findXObjRomOpts(self, rOpts, sizeof(rOpts)/sizeof(XObjRomOpt), 0);
}

static xkern_return_t
ppp_rom_opts(self, strs, nFields, line, arg)
    XObj        self;
    char        **strs;
    int         nFields, line;
    VOID        *arg;
{
  IPhost here, there;
  char * device;
  int mtu;
  PSTATE * pstate = (PSTATE *)self->state;
  IP_LINK * iplink = &(pstate->IP_link);

  if (strcmp(strs[4], self->instName)) {
    if (sscanf(strs[4], "%d", &mtu) < 1) return XK_SUCCESS;
  }

  if (str2ipHost(&here, strs[2]) == XK_FAILURE) {
    xTraceP1(self, TR_ERRORS, "rom parse error: not an IPhost - line %d",line);
    return XK_FAILURE;
  }
  if (str2ipHost(&there, strs[3]) == XK_FAILURE) {
    xTraceP1(self, TR_ERRORS, "rom parse error: not an IPhost - line %d",line);
    return XK_FAILURE;
  }
  if (sscanf(strs[4], "%d", &mtu) == 1) {
    device = 0;
    iplink->port = mtu;
  }
  else {
    device = xMalloc(strlen(strs[4]) + 1);
    strcpy(device, strs[4]);
    iplink->port = 0;
  }
  if (sscanf(strs[5], "%d", &mtu) != 1) {
    xTraceP1(self, TR_ERRORS, "rom parse error: not an MTU - line %d",line);
    return XK_FAILURE;
  }
  iplink->local = here;
  iplink->remote = there;
  iplink->devname = device;
  return XK_SUCCESS;
}

/* ------------------------------------------------------------------------ */
