/*     
 * $RCSfile: psync.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Log: psync.c,v $
 * Revision 1.2  1994/05/06 20:26:16  menze
 *   [ 1994/05/06          menze ]
 *   Fixed older-compiler promotion problem in chk_parents
 *   Use msgShow rather than MSG_SHOW
 *
 * Revision 1.1  1994/04/18  22:02:05  slm
 * Initial revision
 *
 */

/*
 * Abreviations:
 *    Ih = Image of context graph on host h
 *    Vp = participant p's View of context graph
 * Currently, these are the same, since PSYNC always passes up the messages it
 * receives.  (I don't see where it waits for a receive request - slm)
 */

#include "xkernel.h"
#include "psync.h"
#include "psync_internal.h"

#define MSG_SHOW	msgShow

       void           psync_init(XObj);
static void           getproc_protl(XObj), getproc_sessn(XObj);
static XObj           psync_open(XObj, XObj, XObj, Part *);
static xkern_return_t psync_openenable(XObj, XObj, XObj, Part *);
static xkern_return_t psync_demux(XObj, XObj, Msg *);
static int            psync_controlprotl(XObj, int, char *, int);
static xmsg_handle_t  psync_push(XObj, Msg *);
static xkern_return_t psync_pop(XObj, XObj, Msg *, VOID *);
static int            psync_controlsessn(XObj, int, char *, int);
static long           psync_hdr_peek(VOID *, char *, long, VOID *);
static long           psync_hdr_strip(VOID *, char *, long, VOID *);
static void           psync_hdr_store(VOID *, char *, long, VOID *);
static void           node_hdr_store(VOID *, char *, long, VOID *);
static void           rec_own_msg(Event, VOID *);
static void           do_reply(XObj, NODE **, NODE *, PSYNCheader2 *, int);
static void           do_restart(XObj, CID *, NODE *, XObj);
static int            add_node(PSYNCstate *, Sstate *, PSYNCheader2 *, Msg *,
			       unsigned char *, MID *, int, int *);
static int            add_orphan(Sstate *, Msg *);
static void           chk_orphans(PSYNCstate *, XObj);
static int            chk_parents(Sstate *, u_int, MID *, NODE **);
static int            chk_orphan_list(Sstate *, MID *);

/* External declarations of functions defined in psync_support.c */
extern void init_nodes( void );
extern void insert_inviteQ(CID *, PSYNCport, Msg *, XObj);
extern int  init_graph(PSYNCstate *, Sstate *, PSYNCheader1 *, Msg *, int);
extern int  update_graph(PSYNCstate *, Sstate *, MID *);
extern void start_stability_check(PSYNCstate *, Sstate *, MID *);
extern void get_leaves(PSYNCstate *, Sstate *, unsigned char *, MID *);
extern NODE *get_node(PSYNCstate *, Sstate *, MID *);
extern MID  *get_mid(PSYNCstate *, NODE *);
extern Msg  *node_to_msg(PSYNCstate *, NODE *);
extern int  get_sender(PSYNCstate *, Sstate *, MID *);
extern int  check_path(NODE *, NODE *);
extern NODE *newnode(Sstate *, MID *, int, Msg *);
extern void compute_indegrees(NODE **, NODE *, Map *, INDEGREE *);
extern void dumpHdr1(PSYNCheader1 *);
extern void dumpHdr2(PSYNCheader2 *);
extern void dumpHdr3(PSYNCheader3 *);
extern void dumpNodes(NODE *);

#ifdef BROADCAST
IPhost broadcast_addr = { 192, 12, 69, 60 };
#endif

/* Q for participants invited to the conversation but not yet listening. */
INVITEQ *inviteQ = NULL;

int tracepsyncp = 1;

void psync_init(self)
XObj self;
{
    Part       part[1];
    PSYNCstate *pstate;

    xAssert(xIsProtocol(self));
    xTrace0(psyncp, TR_GROSS_EVENTS, "PSYNC init");

    getproc_protl(self);

    if ((pstate = (PSYNCstate *)xMalloc(sizeof(PSYNCstate))) == NULL)
	Kabort("PSYNC: malloc failed in psync_init");
    self->state = (char *)pstate;
    /* Get my IP address from IP. */
    if (xControl(xGetDown(self,0), GETMYHOST, (char *)&pstate->host,
	         sizeof(IPhost)) == -1)
	Kabort("PSYNC: Could not get my IP address");
    pstate->conv_id        = 1;
    pstate->map_openenable = mapCreate(100, sizeof(PSYNCaddr));
    pstate->map_conv       = mapCreate(2*MAX_CONV, sizeof(CID));
    semInit(&pstate->Ipsessn, 1);
    semInit(&pstate->Ihgraph, 1);
    init_nodes();

    partInit(part, 1);
    partPush(part[0], ANY_HOST, 0);

    /* OpenEnable IP. */
    if (xOpenEnable(self, self, xGetDown(self,0), part) == XK_FAILURE) {
        xError("PSYNC: error in init");
        return;
    }
}

static void getproc_protl(self)
XObj self;
{
    xAssert(xIsProtocol(self));

    self->open       = psync_open;
    self->openenable = psync_openenable;
    self->demux      = psync_demux;
    self->control    = psync_controlprotl;
}

static void getproc_sessn(self)
XObj self;
{
    xAssert(xIsSession(self));

    self->push    = psync_push;
    self->pop     = psync_pop;
    self->control = psync_controlsessn;
}

/*
 * Called by hlp to actively establish a psync conversation.  Partlist p
 * consists of list of participant addresses, the 0th one being the active open
 * and the others passive opens.
 */
static XObj psync_open(self, hlpRcv, hlpType, p)
XObj self, hlpRcv, hlpType;
Part *p;	/* p = [psync1,...,psyncN,NULL] */
{
    PSYNCstate *pstate = (PSYNCstate *)self->state;
    XObj       sessn;
    Sstate     *sstate;
    int        i;
    PSYNCaddr  *psyncaddr;
    Part       part[1];
#ifdef BROADCAST
    IPhost     rm_iphost;
#endif

    xAssert(xIsProtocol(self));
    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: entering psync_open");

    /*
     * This is an active open, invoked by the hlpRcv.  Create a PSYNC session
     * (1 per conversation) and set its down session to NULL.
     */
    if (!p || (sessn = xCreateSessn(getproc_sessn, hlpRcv, hlpType, self, 0,
			            (XObj *)NULL)) == ERR_XOBJ)
	return ERR_XOBJ;

    /* Initialize the session state. */
    if ((sstate = (Sstate *)xMalloc(sizeof(Sstate))) == NULL) {
	xError("PSYNC: malloc failed in psync_open");
	return ERR_XOBJ;
    }
    sessn->state = (char *)sstate;
    sstate->first_msg          = TRUE;
    sstate->loopback           = FALSE;
    sstate->root               = NULL;
    sstate->msg_id             = 1;
    sstate->num_orphans        = 0;
    sstate->pushing_mid.msg_id = 0;
    sstate->rec_sessn          = ERR_XOBJ;
    sstate->rec_opcode         = -1;
    sstate->next_msg           = NULL;
    sstate->last_msg           = NULL;
    sstate->down_part_mask     = 0;
    sstate->ignore_part_mask   = 0;
    sstate->map_tree           = mapCreate(2*MAX_NODES, sizeof(MID));
    semInit(&sstate->Pushing, 1);
#ifdef BROADCAST
    /* Open an IP broadcast session. */
    xTrace0(psyncp, 1, "PSYNC: open: openning broadcast IP");
    rm_iphost = broadcast_addr;
    partInit(part, 1);
    partPush(part[0], (VOID *)&rm_iphost, sizeof(IPhost));
    semWait(&pstate->Ipsessn);
    xSetDown(sessn, 0, xOpen(self, self, xGetDown(self,0), part));
    semSignal(&pstate->Ipsessn);
#endif
    /*
     * Open an IP session for each remote host except this one,
     * which already has one.
     */
    for (i = 0; psyncaddr = (PSYNCaddr *)partPop(p[i]); i++) {
	xTrace3(psyncp, TR_DETAILED,
		"PSYNC: part %d, PSYNCport = %d, IPhost = %s",
                i, psyncaddr->port, ipHostStr(&psyncaddr->host));
	sstate->part_list[i] = *psyncaddr;
        if (SAME_PSYNChost(psyncaddr->host, pstate->host)) {
	    /* Participant is on same host; give a new conversation id. */
            sstate->mypartnum   = i;
            sstate->psyncaddr   = *psyncaddr;
            sstate->cid.host    = psyncaddr->host;
            sstate->cid.conv_id = ((PSYNCstate *)self->state)->conv_id++;
            /* `map_conv' is <iphost,id> --> <psync_sessn> */
            mapBind(pstate->map_conv, (VOID *)&sstate->cid, (VOID *)sessn);
        }
        else {
	    /* Process is on remote host; open an IP session to that host. */
            partInit(part, 1);
            partPush(part[0], (VOID *)&psyncaddr->host, sizeof(IPhost));
	    semWait(&pstate->Ipsessn);
	    xSetDown(sessn, i, xOpen(self, self, xGetDown(self,0), part));
	    semSignal(&pstate->Ipsessn);
        }
    }
    sstate->num_part = i;
    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: exiting psync_open");
    return sessn;
}

/*
 * Called by hlp to passively open a conversation.
 */
static xkern_return_t psync_openenable(self, hlpRcv, hlpType, p)
XObj self, hlpRcv, hlpType;
Part *p;	/* p = [psync1] */
{
    PSYNCstate   *pstate = (PSYNCstate *)self->state;
    PSYNCaddr    *psyncaddr;
    INVITEQ      *que, *nextque, *prevque;
    PSYNCheader1 hdr1;
    XObj         sessn;
    Sstate       *sstate;
    Part         part[1];
    int          i;
#ifdef BROADCAST
    IPhost       rm_iphost;
#endif

    xAssert(xIsProtocol(self));
    xAssert(xIsProtocol(hlpRcv));
    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: entering psync_openenable");

    psyncaddr = (PSYNCaddr *)partPop(p[0]);
    xTrace2(psyncp, TR_DETAILED, "PSYNC: PSYNCport = %d, IPhost = %s",
            psyncaddr->port, ipHostStr(&psyncaddr->host));

    /* Check if first message already arrived. */
    prevque = NULL;
    /* Note: msgs added to inviteQ in psync_demux. */
    for (que = inviteQ; que; que = nextque) {
	nextque = que->next;
        if (SAME_PSYNChost(que->cid.host, pstate->host) &&
            que->cid.conv_id == pstate->conv_id &&
	    que->port == psyncaddr->port) {
            /* First message has arrived; create the session. */
            msgPop(que->msg, psync_hdr_peek, (VOID *)&hdr1,
		   sizeof(PSYNCheader1), (VOID *)NULL);
            /* Create a PSYNC session and set its down session to NULL. */
            if ((sessn = xCreateSessn(getproc_sessn, hlpRcv, hlpType, self, 0,
				      (XObj *)NULL)) == ERR_XOBJ) {
	        xError("PSYNC: can't create session in psync_openenable");
	        return XK_FAILURE;
            }
            if ((sstate = (Sstate *)xMalloc(sizeof(Sstate))) == NULL) {
	        xError("PSYNC: malloc failed in psync_openenable");
		return XK_FAILURE;
	    }
            sessn->state = (char *)sstate;
            sstate->first_msg          = TRUE;
            sstate->loopback           = FALSE;
            sstate->root               = NULL;
            sstate->psyncaddr          = *psyncaddr;
            sstate->cid                = hdr1.cid;
            sstate->msg_id             = 1;
            sstate->num_part           = hdr1.num_part;
            sstate->num_orphans        = 0;
            sstate->pushing_mid.msg_id = 0;
            sstate->rec_sessn          = ERR_XOBJ;
            sstate->rec_opcode         = -1;
            sstate->next_msg           = NULL;
            sstate->last_msg           = NULL;
            sstate->down_part_mask     = 0;
            sstate->ignore_part_mask   = 0;
            sstate->map_tree           = mapCreate(2*MAX_NODES, sizeof(MID));
            semInit(&sstate->Pushing, 1);
	    /* Bind the current session to the conversation id. */
            /* `map_conv' is <iphost,id> --> <psync_sessn> */
            mapBind(pstate->map_conv, (VOID *)&sstate->cid, (VOID *)sessn);
            for (i = 0; i <= hdr1.num_part; i++)
	        sstate->part_list[i] = hdr1.part_list[i];
#ifdef BROADCAST
            /* Open an IP broadcast session. */
            xTrace0(psyncp, 1, "PSYNC: openenable: openning broadcast IP");
            rm_iphost = broadcast_addr;
            partInit(part, 1);
            partPush(part[0], (VOID *)&rm_iphost, sizeof(IPhost));
            semWait(&pstate->Ipsessn);
            xSetDown(sessn, 0, xOpen(self, self, xGetDown(self,0), part));
            semSignal(&pstate->Ipsessn);
#endif
	    /*
	     * Open an IP session for each remote host except this one,
	     * which already has one.
	     */
            for (i = 0; i < hdr1.num_part; i++) {
	        if (SAME_PSYNChost(hdr1.part_list[i].host, pstate->host))
	            sstate->mypartnum = i;
	        else {
	            partInit(part, 1);
                    partPush(part[0], (VOID *)&hdr1.part_list[i].host,
                             sizeof(IPhost));
	            semWait(&pstate->Ipsessn);
	            xSetDown(sessn, i,
			     xOpen(self, self, xGetDown(self,0), part));
	            semSignal(&pstate->Ipsessn);
		}
	    }
            if (xOpenDone(hlpRcv, sessn, self) == XK_FAILURE) {
	        xError("PSYNC: error in openenable");
	        return XK_FAILURE;
	    }
            xPop(sessn, que->lls, que->msg, (VOID *)NULL);
	    msgDestroy(que->msg);
	    xFree((char *)que->msg);
	    if (prevque)
	        prevque->next = nextque;
	    else
		inviteQ = nextque;
	    xFree((char *)que);
        }
	else
	    prevque = que;
    }

    /* Bind the passive open info <port,host> into a map. */
    /* `map_openenable' is <localport,iphost> --> <hlp> */
    mapBind(pstate->map_openenable, (VOID *)psyncaddr, (VOID *)hlpRcv);
    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: exiting psync_openenable");
    return XK_SUCCESS;
}

/*
 * Take a message from llp.
 */
static xkern_return_t psync_demux(self, lls, msg)
XObj self, lls;
Msg  *msg;
{
    PSYNCstate	 *pstate = (PSYNCstate *)self->state;
    PSYNCheader3 hdr3;
    XObj         sessn;
    PSYNCheader1 hdr1;
    int          i, j;
    XObj         hlp;
    Sstate       *sstate;
    Part	 part[1];
    Msg          *tmpmsg;
#ifdef BROADCAST
    IPhost       rm_iphost;
#endif

    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: entering psync_demux");

    /* Peek at first two fields in header, type and cid. */
    msgPop(msg, psync_hdr_peek, (VOID *)&hdr3, sizeof(PSYNCheader3),
	   (VOID *)NULL);
    dumpHdr3(&hdr3);
    xIfTrace(psyncp, TR_DETAILED) {
        MSG_SHOW(msg);
    }

    /*
     * Determine receiving conversation.  Note: the message could have
     * PSYNCheader1 or 2, but we don't care, since the members cid and sender
     * are identical for both structures.
     */
    /* `map_conv' is <iphost,id> --> <psync_sessn> */
    if (mapResolve(pstate->map_conv, (VOID *)&hdr3.cid, (VOID **)&sessn) !=
	    XK_FAILURE) {
        /* Pop message to the PSYNC session that runs the conversation. */
        xPop(sessn, lls, msg, (VOID *)NULL);
        xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: exiting psync_demux");
        return XK_SUCCESS;
    }
    /* Probably a conversation initiated by a remote participant. */
    if (hdr3.type == FIRST) {
	xTrace0(psyncp, TR_DETAILED,
		    "PSYNC: first message passing through demux");
        msgPop(msg, psync_hdr_peek, (VOID *)&hdr1, sizeof(PSYNCheader1),
		   (VOID *)NULL);
	dumpHdr1(&hdr1);
        /* Check if the local participant has done an openenable. */
        for (i = 0; i < hdr1.num_part; i++) {
	    if (SAME_PSYNChost(hdr1.part_list[i].host, pstate->host)) {
		/* Determine if hdr1.part_list[i] has done an openenable */
                /* `map_openenable' is <localport,iphost> --> <hlp> */
                if (mapResolve(pstate->map_openenable,
			       (VOID *)&hdr1.part_list[i],
			       (VOID **)&hlp) == XK_FAILURE) {
		    xTrace0(psyncp, TR_DETAILED,
                        "PSYNC: local participant has not done an openenable;");
		    xTrace0(psyncp, TR_DETAILED,
			    "       store the message for future");
		    if ((tmpmsg = (Msg *)xMalloc(sizeof(Msg))) == NULL) {
			xError("PSYNC: malloc failed in psync_demux");
			return XK_FAILURE;
		    }
	            msgConstructCopy(tmpmsg, msg);
	            insert_inviteQ(&hdr1.cid, hdr1.part_list[i].port, tmpmsg,
				   lls);
	        }
	        else {
		    xTrace0(psyncp, TR_DETAILED,
			    "PSYNC: create a PSYNC session");
		    /* Create a PSYNC session & set its down session to lls. */
      	            if ((sessn = xCreateSessn(getproc_sessn, hlp, hlp, self, 1,
					      &lls)) == ERR_XOBJ) {
			xError("PSYNC: can't create session in psync_demux");
			return XK_FAILURE;
		    }
	            /* Init the session state. */
	            if ((sstate = (Sstate *)xMalloc(sizeof(Sstate))) == NULL) {
	                xError("PSYNC: malloc failed in psync_demux");
			return XK_FAILURE;
		    }
	            sessn->state = (char *)sstate;
	            sstate->first_msg          = TRUE;
	            sstate->loopback           = FALSE;
	            sstate->root               = NULL;
	            sstate->mypartnum          = i;
	            sstate->psyncaddr          = hdr1.part_list[i];
	            sstate->cid                = hdr1.cid;
	            sstate->msg_id             = 1;
	            sstate->num_part           = hdr1.num_part;
	            sstate->num_orphans        = 0;
	            sstate->pushing_mid.msg_id = 0;
                    sstate->rec_sessn          = ERR_XOBJ;
                    sstate->rec_opcode         = -1;
                    sstate->next_msg           = NULL;
                    sstate->last_msg           = NULL;
                    sstate->down_part_mask     = 0;
                    sstate->ignore_part_mask   = 0;
	            sstate->map_tree = mapCreate(2*MAX_NODES, sizeof(MID));
	            semInit(&sstate->Pushing, 1);
	            /* Bind the current sessn to the conversation id. */
                    /* `map_conv' is <iphost,id> --> <psync_sessn> */
                    mapBind(pstate->map_conv, (VOID *)&sstate->cid,
			    (VOID *)sessn);
	            for (j = 0; j <= hdr1.num_part; j++)
	                sstate->part_list[j] = hdr1.part_list[j];
#ifdef BROADCAST
	            /* Open an IP broadcast session. */
	            xTrace0(psyncp, 1, "PSYNC: demux: openning broadcast IP");
                    rm_iphost = broadcast_addr;
	            partInit(part, 1);
                    partPush(part[0], (VOID *)&rm_iphost, sizeof(IPhost));
                    semWait(&pstate->Ipsessn);
                    xSetDown(sessn, 0,
			     xOpen(self, self, xGetDown(self,0), part));
                    semSignal(&pstate->Ipsessn);
#endif
		    /*
		     * Open an IP session for each remote host except first,
		     * which already has one.
		     */
	            for (j = 1; j < hdr1.num_part; j++) {
	                if (j != i) {
                            partInit(part, 1);
                            partPush(part[0], (VOID *)&hdr1.part_list[j].host,
                                     sizeof(IPhost));
                            semWait(&pstate->Ipsessn);
                            xSetDown(sessn, j,
				     xOpen(self, self, xGetDown(self,0), part));
                            semSignal(&pstate->Ipsessn);
	                }
		    }
                    if (xOpenDone(hlp, sessn, hlp) == XK_FAILURE) {
        		xError("PSYNC: error in demux");
        		return XK_FAILURE;
        	    }
                    xPop(sessn, lls, msg, (VOID *)NULL);
	        }
	        break;
	    }
	}
    }
    else {
        /*
	* This is not the first message of conversation and local
	* participant hasn't done any openenable, so ignore the message.
        */
        xTrace0(psyncp, TR_DETAILED, "PSYNC: demux: dropping message");
    }
    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: exiting psync_demux");
    return XK_SUCCESS;
}

static int psync_controlprotl(self, opcode, buf, len)
XObj self;
int  opcode, len;
char *buf;
{
    xAssert(xIsProtocol(self));

    Kabort("PSYNC: psync_controlprotl() not supported");

    switch (opcode) {
	/* where is PSYNC_RESTORESESSN ??? - slm */

        default:
            return -1;
    }
}

/*
 * Send a message message to llp for every remote host.
 */
static xmsg_handle_t psync_push(self, msg)
XObj self;
Msg  *msg;
{
    Sstate        *sstate = (Sstate *)self->state;
    PSYNCstate    *pstate = (PSYNCstate *)xMyProtl(self)->state;
    ARGstruct     *arg;
    Msg           *dg, msgsend;
    PART_MASK     mask;
    int           i, msg_node = 0, status;
    PSYNCheader1  hdr1;
    PSYNCheader2  hdr2;
    unsigned char num_depend;
    MID           depend_list[MAX_PART];

    xAssert(xIsSession(self));

    xTrace3(psyncp, TR_FUNCTIONAL_TRACE,
	    "PSYNC: entering psync_push, self = %d, msg = %x, len = %d",
	    self, msg, msgLen(msg));
    if ((arg = (ARGstruct *)xMalloc(sizeof(ARGstruct))) == NULL ||
        (dg  = (Msg *)xMalloc(sizeof(Msg))) == NULL) {
	xError("PSYNC: malloc failed in psync_push");
	return XMSG_ERR_HANDLE;
    }

    /* Push a msg only if I am alive. */
    mask = 1 << sstate->mypartnum;
    if (sstate->down_part_mask & mask) {
        xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: exiting psync_push");
	return 0;
    }

    /*
     * Since multiple processes may be trying to push, lock the push before
     * pushing.  The lock is released either here or when the message being
     * pushed is acknowledged in control_sessn.
     */
    xTrace0(psyncp, TR_DETAILED, "PSYNC: waiting on push sem");
    semWait(&sstate->Pushing);
    xTrace0(psyncp, TR_DETAILED, "PSYNC: done waiting on push sem");

    /* Make a copy of the msg. */
    msgConstructCopy(dg, msg);
    if (sstate->first_msg) {	/* first message of the conversation */
        /*
	 * Add a header of type PSYNCheader1 and push the msg to each remote
	 * participant.
         */
	bzero((char *)&hdr1, sizeof(PSYNCheader1));
        hdr1.type       = FIRST;
        hdr1.cid        = sstate->cid;
        hdr1.mid.host   = pstate->host;
	hdr1.mid.msg_id = sstate->msg_id++;
        hdr1.sender     = sstate->mypartnum;
        hdr1.num_part   = sstate->num_part;
        for (i = 0; i < hdr1.num_part; i++)
            hdr1.part_list[i] = sstate->part_list[i];
        msgPush(msg, psync_hdr_store, (VOID *)&hdr1, sizeof(PSYNCheader1),
		(VOID *)NULL);
        sstate->pushing_mid = hdr1.mid;
        xTrace1(psyncp, TR_FUNCTIONAL_TRACE,
		"PSYNC: pushing first msg of len %d", msgLen(msg));
        dumpHdr1(&hdr1);
        xIfTrace(psyncp, TR_DETAILED) {
	    MSG_SHOW(msg);
	}
#ifdef BROADCAST
        msgConstructCopy(&msgsend, msg);
        xPush(xGetDown(self,0), msgsend);
	msgDestroy(&msgsend);
#else
        /* Push message down to llp for every remote host. */
        for (i = 0, mask = 1; i < sstate->num_part; mask <<= 1, i++) {
            if (i != sstate->mypartnum && !(sstate->down_part_mask & mask)){
  	        if (xIsXObj(xGetDown(self,i))) {
                    xTrace2(psyncp, TR_DETAILED, "PSYNC: %d pushing to %d ",
			    sstate->mypartnum, i);
		    msgConstructCopy(&msgsend, msg);
	            xPush(xGetDown(self,i), &msgsend);
		    msgDestroy(&msgsend);
	        }
            }
	}
#endif
        /* This is the first msg, so we have to init the context graph. */
        msg_node = init_graph(pstate, sstate, &hdr1, dg, 1);
        if (sstate->loopback) {
            arg->psync_sessn = self;
            arg->state       = sstate;
            if ((arg->msg    = (Msg *)xMalloc(sizeof(Msg))) == NULL) {
		xError("PSYNC: malloc failed in psync_push");
		return XMSG_ERR_HANDLE;
	    }
	    msgConstructCopy(arg->msg, dg);
            arg->mid         = hdr1.mid;
            /* Schedule thread which sends a copy of this msg to me. */
            evDetach(evSchedule(rec_own_msg, (VOID *)arg, 0));
	    /* semaphore will be released when DIVIDER's demux ACKs the msg */
        }
        else {
	    if (update_graph(pstate, sstate, &hdr1.mid))
                start_stability_check(pstate, sstate, &hdr1.mid);
            semSignal(&sstate->Pushing);
        }
    }
    else {	/* NOT the first message of the conversation */
        /*
	 * Add a header of type PSYNCheader2 and push the msg to every remote
	 * participant.
         */
	bzero((char *)&hdr2, sizeof(PSYNCheader2));
        hdr2.type       = REGULAR;
        hdr2.cid        = sstate->cid;
        hdr2.mid.host   = pstate->host;
	hdr2.mid.msg_id = sstate->msg_id++;
        hdr2.sender     = sstate->mypartnum;
        get_leaves(pstate, sstate, &hdr2.num_depend, hdr2.depend_list);
        msgPush(msg, psync_hdr_store, (VOID *)&hdr2, sizeof(PSYNCheader2),
		(VOID *)NULL);
        sstate->pushing_mid = hdr2.mid;
        xTrace1(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: pushing msg of len %d",
		msgLen(msg));
        dumpHdr2(&hdr2);
        xIfTrace(psyncp, TR_DETAILED) {
	    MSG_SHOW(msg);
        }
#ifdef BROADCAST
        msgConstructCopy(&msgsend, msg);
        xPush(xGetDown(self,0), &msgsend);
	msgDestroy(&msgsend);
#else
	/* Push message down to llp for every remote host. */
        for (i = 0, mask = 1; i < sstate->num_part; mask <<= 1, i++) {
            if (i != sstate->mypartnum && !(sstate->down_part_mask & mask)) {
	        if (xIsXObj(xGetDown(self,i))) {
                    xTrace2(psyncp, TR_DETAILED, "PSYNC: %d pushing to %d ",
		            sstate->mypartnum, i);
		    msgConstructCopy(&msgsend, msg);
printf("about to call IP's xPush\n");
	            xPush(xGetDown(self,i), &msgsend);
printf("returned from IP's xPush\n");
		    msgDestroy(&msgsend);
	        }
            }
	}
#endif
        /* Add the msg to the context graph.  Update Ih. */
        msg_node = add_node(pstate, sstate, &hdr2, dg, &num_depend,
			    depend_list, 1, &status);
        if (status) {
            xTrace1(psyncp, TR_DETAILED,
	            "PSYNC: psync_push: add_node returned %d", status);
	    xTrace2(psyncp, TR_DETAILED, "mid.msg_id = %d, mid.host = %s",
		    hdr2.mid.msg_id, ipHostStr(&hdr2.mid.host));
	    return XMSG_ERR_HANDLE;
        }
        if (sstate->loopback) {
            arg->psync_sessn = self;
            arg->state       = sstate;
            if ((arg->msg    = (Msg *)xMalloc(sizeof(Msg))) == NULL) {
		xError("PSYNC: malloc failed in psync_push");
		return XMSG_ERR_HANDLE;
	    }
	    msgConstructCopy(arg->msg, dg);
            arg->mid         = hdr2.mid;
            /* Schedule thread which sends a copy of this msg to me. */
            evDetach(evSchedule(rec_own_msg, (VOID *)arg, 0));
	    /* semaphore will be released when TOTAL_ORDER's pop ACKs the msg */
        }
        else {
            if (update_graph(pstate, sstate, &hdr2.mid))
                start_stability_check(pstate, sstate, &hdr2.mid);
            semSignal(&sstate->Pushing);
        }
    }
    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: exiting psync_push");
    if (msg_node)
        return msg_node;
    else
	return XMSG_ERR_HANDLE;
}

/*
 * Receive message from remote participant.
 */
static xkern_return_t psync_pop(self, lls, msg, hdrp)
XObj self, lls;
Msg  *msg;
VOID *hdrp;
{
    Sstate        *sstate = (Sstate *)self->state;
    PSYNCstate    *pstate = (PSYNCstate *)xMyProtl(self)->state;
    PSYNCheader1  hdr1;
    PSYNCheader2  hdr2, newhdr2;
    PSYNCheader3  hdr3;
    NODE          *node, *depend[MAX_PART];
    Msg           *dg, *msgcpy, newmsg;
    PART_MASK     mask;
    ARGstruct     arg;
    unsigned char num_depend;
    MID           depend_list[MAX_PART];
    int           status, i, j, count, retval = 0;
    /* retval = 1 means free dg; retval = 2 means free msgcpy */

    xAssert(xIsSession(self));
    xAssert(xIsSession(lls));
    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: entering psync_pop");

    /* Peek into the PSYNCheader3. */
    msgPop(msg, psync_hdr_peek, (VOID *)&hdr3, sizeof(PSYNCheader3),
	   (VOID *)NULL);

    switch (hdr3.type) {
        case FIRST:	/* This is the FIRST message of the conversation. */
            /*
             * Strip off the header, initialize the context graph, add a
	     * node header, check if any orphans have their parents now, and
	     * demux the msg.
             */
            msgPop(msg, psync_hdr_strip, (VOID *)&hdr1, sizeof(PSYNCheader1),
		   (VOID *)NULL);
            dumpHdr1(&hdr1);
            xIfTrace(psyncp, TR_DETAILED) {
                MSG_SHOW(msg);
            }
            if ((dg = (Msg *)xMalloc(sizeof(Msg))) == NULL) {
	        xError("PSYNC: malloc failed in psync_pop");
	        return XK_FAILURE;
            }
            msgConstructCopy(dg, msg);	/* since msg might be destroyed */
            (void)init_graph(pstate, sstate, &hdr1, dg, 0);
            arg.state = sstate;
            arg.mid   = hdr1.mid;
            msgPush(msg, node_hdr_store, (VOID *)&arg, sizeof(NODE *),
		    (VOID *)&arg);
            if (!sstate->loopback) {
		if (update_graph(pstate, sstate, &hdr1.mid))
                    start_stability_check(pstate, sstate, &hdr1.mid);
	    }
            chk_orphans(pstate, self);
            xDemux(self, msg);
            break;

        case REGULAR:
            if ((msgcpy = (Msg *)xMalloc(sizeof(Msg))) == NULL ||
                (dg = (Msg *)xMalloc(sizeof(Msg))) == NULL) {
	        xError("PSYNC: malloc failed in psync_pop");
	        return XK_FAILURE;
            }
	    msgConstructCopy(msgcpy, msg);
            /* Strip off the header. */
            msgPop(msg, psync_hdr_strip, (VOID *)&hdr2, sizeof(PSYNCheader2),
		   (VOID *)NULL);
            dumpHdr2(&hdr2);
            xIfTrace(psyncp, TR_DETAILED) {
                MSG_SHOW(msg);
            }
            mask = 1 << hdr2.sender;
            if ((mask & sstate->down_part_mask) ||
		(mask & sstate->ignore_part_mask))
                break;
            msgConstructCopy(dg, msg);
            if (sstate->root)	/* Ih root already exits; add the current msg
				 * to the context graph */
                (void)add_node(pstate, sstate, &hdr2, dg, &num_depend,
			       depend_list, 0, &status);
            else /* Ih root doesn't exist; first message hasn't been received */
                status = 1;
	    if (status == 0) {	/* message's parents are known in Ih */
  	        /*
		 * Add a node header to the msg, check if any orphans have
		 * their parents now, and demux the msg.
		 */
                arg.state = sstate;
                arg.mid   = hdr2.mid;
                msgPush(msg, node_hdr_store, (VOID *)&arg, sizeof(NODE *),
			(VOID *)&arg);
	        if (!sstate->loopback) {
		    if (update_graph(pstate, sstate, &hdr2.mid))
	                start_stability_check(pstate, sstate, &hdr2.mid);
		}
                chk_orphans(pstate, self);
                xDemux(self, msg);
		retval = 2;
            }
            else if (status > 0) {	/* some parents are missing */
                xTrace2(psyncp, TR_DETAILED,
                    "PSYNC: received an orphan, mid.msg_id = %d, mid.host = %s",
                        hdr2.mid.msg_id, ipHostStr(&hdr2.mid.host));
                /* Add msg to orphan list and prepare a RETRANSMIT request. */
                if (add_orphan(sstate, msgcpy)) {
                    if (status == 1) {	/* retransmit required */
			bzero((char *)&newhdr2, sizeof(PSYNCheader2));
	                newhdr2.type       = RETRANSMIT;
	                newhdr2.cid        = hdr2.cid;
	                newhdr2.mid        = hdr2.mid;
	                newhdr2.sender     = sstate->mypartnum;
			newhdr2.num_depend = num_depend;
			for (i = 1; i < num_depend; i++)
	                    newhdr2.depend_list[i] = depend_list[i];
                        msgConstructBuffer(&newmsg, (char *)&newhdr2,
					   sizeof(PSYNCheader2));
                        xTrace2(psyncp, TR_FUNCTIONAL_TRACE,
                     "PSYNC: pop(REGULAR): RETRANSMIT req - %d pushing to %d",
                                sstate->mypartnum, hdr2.sender);
	                xTrace2(psyncp, TR_FUNCTIONAL_TRACE,
                     "                     request for MID: host = %s, id = %d",
                                ipHostStr(&hdr2.mid.host), hdr2.mid.msg_id);
	                xPush(xGetDown(self,hdr2.sender), &newmsg);
			msgDestroy(&newmsg);
	            }
		}
		else
		    return XK_FAILURE;
		retval = 1;
            }
	    else	/* status == -1 */
		retval = 1;
            break;

        case REPLY:	/* This message is a reply to a RETRANSMIT request. */
            /*
	     * Make a copy of the msg, strip the header, and add the msg to the
	     * context graph.  If some parents are still missing add it to the
	     * orphan list.
             */
            if ((msgcpy = (Msg *)xMalloc(sizeof(Msg))) == NULL ||
                (dg = (Msg *)xMalloc(sizeof(Msg))) == NULL) {
	        xError("PSYNC: malloc failed in psync_pop");
	        return XK_FAILURE;
            }
	    msgConstructCopy(msgcpy, msg);
            msgPop(msg, psync_hdr_strip, (VOID *)&hdr2, sizeof(PSYNCheader2),
		   (VOID *)NULL);
            dumpHdr2(&hdr2);
            xIfTrace(psyncp, TR_DETAILED) {
                MSG_SHOW(msg);
            }
            msgConstructCopy(dg, msg);
            (void)add_node(pstate, sstate, &hdr2, dg, &num_depend, depend_list,
			   0, &status);
	    if (status == 0) {
  	        /*
		 * Add a node header to the msg, check if any orphans have
		 * their parents now, and demux the msg.
		 */
                arg.state = sstate;
                arg.mid   = hdr2.mid;
                msgPush(msg, node_hdr_store, (VOID *)&arg, sizeof(NODE *),
                        (VOID *)&arg);
	        if (!sstate->loopback) {
		    if (update_graph(pstate, sstate, &hdr2.mid))
	                start_stability_check(pstate, sstate, &hdr2.mid);
		}
	        chk_orphans(pstate, self);
                xDemux(self, msg);
		retval = 2;
            }
            else if (status > 0) {
                /*
                 * Some of message's parents are still missing (this is alright,
		 * since we can't expect the messages that were sent in response
		 * to our RETRANSMIT request to be in the right order).
                 */
                if (!add_orphan(sstate, msgcpy))
		    return XK_FAILURE;
		retval = 1;
            }
	    else	/* status == -1 */
		retval = 1;
            break;

        case RETRANSMIT:	/* retransmit request */
            /*
	     * For now, we assume the 2 host test case and therefore only
             * retransmit immediate parents.  Note: when we add code to
             * retransmit the complete subtree, then we also have to add code
             * to add_node() to update the Ih leaf pointers, otherwise we
             * retransmit the whole Ih every time! - shiv
	     */
            /*
	     * If the first mid in depend_list is same as the orphan mid, then
             * the message corresponding to orphan is also sent.
	     */
            msgPop(msg, psync_hdr_strip, (VOID *)&hdr2, sizeof(PSYNCheader2),
		   (VOID *)NULL);
            dumpHdr2(&hdr2);
            xIfTrace(psyncp, TR_DETAILED) {
                MSG_SHOW(msg);
            }
	    /* `map_tree' is <iphost,mid> --> <node> */
            if (mapResolve(sstate->map_tree, (VOID *)&hdr2.mid,
			   (VOID **)&node) == XK_FAILURE) {
                xError("PSYNC: pop(RETRANSMIT): don't know orphan");
                return XK_FAILURE;
            }
            count = 0;
            /* Send REPLY with all the messages preceding this msg. */
            if (SAME_MID(hdr2.mid, hdr2.depend_list[0])) {
                for (i = 1; i < hdr2.num_depend + 1; i++) {
	            if ((depend[count] = get_node(pstate, sstate,
						&hdr2.depend_list[i])) != NULL)
	                count++;
                }
                depend[count] = NULL;
            }
            else {
                for (i = 0; i < hdr2.num_depend; i++) {
	            if ((depend[count] = get_node(pstate, sstate,
                                                &hdr2.depend_list[i])) != NULL)
	                count++;
                }
                depend[count] = NULL;
            }
            if (SAME_MID(hdr2.mid, hdr2.depend_list[0]))
                do_reply(self, depend, node, &hdr2, 1);
            else
                do_reply(self, depend, node, &hdr2, 0);
            break;

        case RESTART:
            Kabort("PSYNC: psync_pop - RESTART msg not supported");
            for (j = 0; j < sstate->num_part; j++)
                if (SAME_PSYNChost(sstate->part_list[j].host, hdr3.cid.host))
                    break;
            for (i = 0; i < sstate->num_leaves; i++)
                do_restart(self, &hdr3.cid, sstate->leaves[i],
			   xGetDown(self,j));
            /*
	     * Set the proper variable to inform the application that the
	     * participant is up.
	     */
            if (sstate->rec_opcode != -1)
/* DOUBTFUL - what does shiv want to do here */
                xControl(sstate->rec_sessn, sstate->rec_opcode,
			 (char *)&hdr3.cid, 10);
            else {
                xError("PSYNC: Recovery protocol not known");
		return XK_FAILURE;
	    }
            break;

        default:
	    xError("PSYNC: unknown message type in psync_pop");
	    return XK_FAILURE;
    }			/* end switch */
    if (sstate->root) {
        xTrace0(psyncp, TR_DETAILED, "PSYNC: dump context graph");
        xIfTrace(psyncp, TR_DETAILED) {
            dumpNodes(sstate->root);
	}
    }
    else
        xTrace0(psyncp, TR_DETAILED, "NULL CONTEXT GRAPH");
    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: exiting psync_pop");
    if (retval == 1) {
	msgDestroy(dg);
	xFree((char *)dg);
    }
    else if (retval == 2) {
	msgDestroy(msgcpy);
	xFree((char *)msgcpy);
    }
    return XK_SUCCESS;
}

/*
 * Provide protocol-specific services to the hlp.
 */
static int psync_controlsessn(self, opcode, buf, len)
XObj self;
int  opcode, len;
char *buf;
{
    Sstate        *sstate = (Sstate *)self->state;
    XObj          myprotl = xMyProtl(self);
    PSYNCstate    *pstate = (PSYNCstate *)myprotl->state;
    NODE          *node, *node2;
    int           i, j, ptr, num;
    PART_MASK     mask;
    MID           *mid;
    PSYNCaddr     psyncaddr;
    Part          part[2];

    switch (opcode) {
	/* Set the LoopBack flag so the sender receives a 'cc' of the msg. */
        case PSYNC_LOOPBACK:	/* used by COMPOSER */
	    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: in PSYNC_LOOPBACK");
            sstate->loopback = TRUE;
            return 0;

        case PSYNC_ACKMESSAGE:	/* used by TOTAL_ORDER & DIVIDER */
	    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: in PSYNC_ACKMESSAGE");
	    checkLen(len, sizeof(NODE *));
            bcopy(buf, (char *)&node, sizeof(NODE *));
            mid = get_mid(pstate, node);
            if (update_graph(pstate, sstate, mid))
                start_stability_check(pstate, sstate, mid);
            if (sstate->pushing_mid.msg_id &&
		SAME_MID(sstate->pushing_mid, *mid)) {
	        /* Acknowledging own message, so release the semaphore. */
	        sstate->pushing_mid.msg_id = 0;
	        semSignal(&sstate->Pushing);
            }
	    xFree((char *)mid);
            return 0;

	/* Return the root of the context graph. */
        case PSYNC_FINDROOT:
	    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: in PSYNC_FINDROOT");
	    checkLen(len, sizeof(NODE *));
	    bcopy((char *)&sstate->root, buf, sizeof(NODE *));
	    return sizeof(NODE *);

	/* Return the current leaves in the context graph. */
        case PSYNC_GETLEAVES:
	    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: in PSYNC_GETLEAVES");
	    checkLen(len, (MAX_PART+1) * sizeof(int));
	    semWait(&pstate->Ihgraph);
	    num = (int)sstate->num_leaves;
            bcopy((char *)&num, buf, sizeof(int));
            bcopy((char *)sstate->leaves, buf+sizeof(int), num*sizeof(int));
	    semSignal(&pstate->Ihgraph);
            return ((num+1) * sizeof(int));

	/* Return the process that sent the node. */
        case PSYNC_SENDER:	/* used by TOTAL_ORDER */
	    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: in PSYNC_SENDER");
	    checkLen(len, MAX(sizeof(NODE *), sizeof(PSYNCaddr)));
            bcopy(buf, (char *)&node, sizeof(NODE *));
            mid = get_mid(pstate, node);
            psyncaddr = sstate->part_list[get_sender(pstate, sstate, mid)];
            bcopy((char *)&psyncaddr, buf, sizeof(PSYNCaddr));
	    xFree((char *)mid);
            return sizeof(PSYNCaddr);

	/* Same as GETPARTICIPANTS in other protocols. */
        case PSYNC_GETPARTS:	/* used by TOTAL_ORDER */
	    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: in PSYNC_GETPARTS");
	    checkLen(len, sizeof(char) + MAX_PART*sizeof(PSYNCaddr));
            bcopy((char *)&sstate->num_part, buf, sizeof(char));
            bcopy((char *)sstate->part_list, buf+sizeof(char),
		  MAX_PART*sizeof(PSYNCaddr));
            return (sizeof(char) + MAX_PART*sizeof(PSYNCaddr));

	/* Return the set of nodes directly following this node in the context
	 * graph (the set of nodes that depend on this node). */
        case PSYNC_NEXTNODES:	/* used by TOTAL_ORDER */
	    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: in PSYNC_NEXTNODES");
	    checkLen(len, (MAX_PART+1) * sizeof(int));
            bcopy(buf, (char *)&node, sizeof(NODE *));
	    semWait(&pstate->Ihgraph);
	    num = (int)node->num_next;
            bcopy((char *)&num, buf, sizeof(int));
	    bcopy((char *)node->next, buf+sizeof(int), num*sizeof(NODE *));
	    semSignal(&pstate->Ihgraph);
            return (sizeof(int) + num*sizeof(NODE *));

	/* Return the set of nodes directly preceding this node in the context
	 * graph (the set of nodes on which this node depends). */
        case PSYNC_PREVNODES:
	    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: in PSYNC_PREVNODES");
	    checkLen(len, sizeof(int) + MAX_PART*sizeof(NODE *));
            bcopy(buf, (char *)&node, sizeof(NODE *));
	    semWait(&pstate->Ihgraph);
	    num = (int)node->num_prev;
            bcopy((char *)&num, buf, sizeof(int));
            bcopy((char *)node->prev, buf+sizeof(int), num*sizeof(NODE *));
	    semSignal(&pstate->Ihgraph);
            return (sizeof(int) + num*sizeof(NODE *));

	/* Return TRUE if there are outstanding nodes. */
        case PSYNC_OUTSTANDING:
	    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: in PSYNC_OUTSTANDING");
	    checkLen(len, sizeof(int));
            if (sstate->next_msg)
		buf[0] = TRUE;
	    else
		buf[0] = FALSE;
            return 1;

	/* Return TRUE if node precedes node2. */
        case PSYNC_PRECEDES:	/* used by TOTAL_ORDER & MEMBERSHIP */
	    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: in PSYNC_PRECEDES");
	    checkLen(len, 2*sizeof(NODE *));
            bcopy(buf, (char *)&node, sizeof(NODE *));
            bcopy(buf+sizeof(NODE *), (char *)&node2, sizeof(NODE *));
	    semWait(&pstate->Ihgraph);
            if (check_path(node, node2))
		buf[0] = TRUE;
	    else
		buf[0] = FALSE;
	    semSignal(&pstate->Ihgraph);
            return 1;

	/* Return TRUE if node is stable. */
        case PSYNC_STABLE:	/* used by TOTAL_ORDER */
	    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: in PSYNC_STABLE");
	    checkLen(len, sizeof(int));
            bcopy(buf, (char *)&node, sizeof(NODE *));
            if (node->stable_part_mask == 0)
	        buf[0] = TRUE;
	    else
		buf[0] = FALSE;
            return 1;

        case PSYNC_UNSTABLE:	/* used by MEMBERSHIP & MONITOR */
	    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: in PSYNC_UNSTABLE");
	    checkLen(len, sizeof(int) + MAX_PART*sizeof(PSYNCaddr));
            bcopy(buf, (char *)&node, sizeof(NODE *));
            if (node->stable_part_mask == 0)
		return 0;
            num = 0;
            ptr = sizeof(int);
            for (i = 0, mask = 1; i < sstate->num_part; i++, mask <<= 1) {
	        if (((node->stable_part_mask & mask) &
		    ~(sstate->down_part_mask))) {
	            bcopy((char *)&sstate->part_list[i], buf+ptr,
			  sizeof(PSYNCaddr));
	            ptr += sizeof(PSYNCaddr);
	            num++;
	        }
	    }
            bcopy((char *)&num, buf, sizeof(int));
            return ptr;

        case PSYNC_MASKOUT:	/* used by MEMBERSHIP & RECOVERY */
	    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: in PSYNC_MASKOUT");
	    checkLen(len, sizeof(PSYNCaddr));
            bcopy(buf, (char *)&psyncaddr, sizeof(PSYNCaddr));
            for (i = 0, mask = 1; i < sstate->num_part; i++, mask <<= 1) {
                if (SAME_PSYNChost(psyncaddr.host, sstate->part_list[i].host)) {
	            sstate->down_part_mask |= mask;
	            for (j = 0; j < sstate->num_leaves; j++) {
			mid = get_mid(pstate, sstate->leaves[j]);
	                start_stability_check(pstate, sstate, mid);
			xFree((char *)mid);
		    }
	            break;
                }
	    }
            return 0;

        case PSYNC_MASKIN:	/* used by MEMBERSHIP */
	    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: in PSYNC_MASKIN");
	    checkLen(len, sizeof(PSYNCaddr));
            bcopy(buf, (char *)&psyncaddr, sizeof(PSYNCaddr));
            for (i = 0, mask = 1; i < sstate->num_part; i++, mask <<= 1)
                if (SAME_PSYNChost(psyncaddr.host, sstate->part_list[i].host)) {
	            if (i != sstate->mypartnum && !xIsXObj(xGetDown(self,i))) {
	                partInit(part, 2);
	                partPush(part[0], (VOID *)&pstate->host, sizeof(int));
                        partPush(part[1], (VOID *)&sstate->part_list[i].host,
				 sizeof(IPhost));
	                semWait(&pstate->Ipsessn);
	                xSetDown(self, i,
		            xOpen(myprotl, myprotl, xGetDown(myprotl,0), part));
	                semSignal(&pstate->Ipsessn);
	            }
	            sstate->down_part_mask &= ~mask;
	            sstate->ignore_part_mask &= ~mask;
	            break;
                }
            return 0;

	/* where is PSYNC_RSTART ??? - slm */

	/* Return the handle to the msg in the context graph. */
        case PSYNC_GETHANDLE:
	    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: in PSYNC_GETHANDLE");
	    checkLen(len, MAX(sizeof(MID), sizeof(NODE *)));
	    if ((mid = (MID *)xMalloc(sizeof(MID))) == NULL) {
		xError("PSYNC: malloc failed in psync_controlsessn");
		return -1;
	    }
            bcopy(buf, (char *)mid, sizeof(MID));
            if (node = get_node(pstate, sstate, mid))
	        bcopy((char *)&node, buf, sizeof(NODE *));
	    else
		bzero(buf, sizeof(NODE *));
	    xFree((char *)mid);
	    return sizeof(NODE *);

	/* Given the msg handle, return the mid of the msg. */
        case PSYNC_GETMID:
	    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: in PSYNC_GETMID");
	    checkLen(len, MAX(sizeof(NODE *), sizeof(MID)));
            bcopy(buf, (char *)&node, sizeof(NODE *));
            bcopy((char *)&node->mid, buf, sizeof(MID));
            return sizeof(MID);

        case PSYNC_ISALIVE:	/* used by DISPATCH & MONITOR */
	    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: in PSYNC_ISALIVE");
	    checkLen(len, sizeof(PSYNCaddr));
            bcopy(buf, (char *)&psyncaddr, sizeof(PSYNCaddr));
            for (i = 0, mask = 1; i < sstate->num_part; i++, mask <<= 1) {
                if (SAME_PSYNChost(psyncaddr.host, sstate->part_list[i].host)) {
	            if (sstate->down_part_mask & mask)
	                buf[0] = FALSE;
	            else
	                buf[0] = TRUE;
		    break;
                }
	    }
	    return 1;

        case PSYNC_LASTMSG:	/* used by MEMBERSHIP */
	    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: in PSYNC_LASTMSG");
	    checkLen(len, MAX(sizeof(PSYNCaddr), sizeof(NODE *)));
            bcopy(buf, (char *)&psyncaddr, sizeof(PSYNCaddr));
            for (i = 0; i < sstate->num_part; i++)
                if (SAME_PSYNChost(psyncaddr.host, sstate->part_list[i].host)) {
	            bcopy((char *)&sstate->lastmsg[i], buf, sizeof(NODE *));
		    break;
		}
            return sizeof(NODE *);

        case PSYNC_IGNORE:	/* used by MEMBERSHIP */
	    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: in PSYNC_IGNORE");
	    checkLen(len, sizeof(PSYNCaddr));
            bcopy(buf, (char *)&psyncaddr, sizeof(PSYNCaddr));
            for (i = 0, mask = 1; i < sstate->num_part; i++, mask <<= 1)
                if (SAME_PSYNChost(psyncaddr.host, sstate->part_list[i].host)) {
	            sstate->ignore_part_mask |= mask;
	            buf[0] = TRUE;
		    return 1;
                }
	    buf[0] = FALSE;
            return 1;

        case PSYNC_ISIGNORE:	/* used by MONITOR */
	    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: in PSYNC_ISIGNORE");
	    checkLen(len, sizeof(PSYNCaddr));
            bcopy(buf, (char *)&psyncaddr, sizeof(PSYNCaddr));
            for (i = 0, mask = 1; i < sstate->num_part; i++, mask <<= 1) {
                if (SAME_PSYNChost(psyncaddr.host, sstate->part_list[i].host)) {
	            if (sstate->ignore_part_mask & mask)
	                buf[0] = TRUE;
	            else
	                buf[0] = FALSE;
		    break;
                }
	    }
	    return 1;

        case PSYNC_ACCEPT:	/* used by MEMBERSHIP */
	    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: in PSYNC_ACCEPT");
	    checkLen(len, sizeof(PSYNCaddr));
            bcopy(buf, (char *)&psyncaddr, sizeof(PSYNCaddr));
            for (i = 0, mask = 1; i < sstate->num_part; i++, mask <<= 1)
                if (SAME_PSYNChost(psyncaddr.host, sstate->part_list[i].host)) {
	            sstate->ignore_part_mask &= ~mask;
		    buf[0] = TRUE;
		    return 1;
                }
	    buf[0] = FALSE;
            return 1;

        case PSYNC_SETRECOVERY:	/* used by RECOVERY */
	    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: in PSYNC_SETRECOVERY");
	    checkLen(len, sizeof(XObj)+sizeof(int));
            bcopy(buf, (char *)&sstate->rec_sessn, sizeof(XObj));
            bcopy(buf+sizeof(XObj), (char *)&sstate->rec_opcode, sizeof(int));
            return 0;

        case PSYNC_NODETOMSG:
	    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: in PSYNC_NODETOMSG");
	    checkLen(len, MAX(sizeof(NODE *), sizeof(Msg *)));
            bcopy(buf, (char *)&node, sizeof(NODE *));
            num = (int)node_to_msg(pstate, node);
	    bcopy((char *)&num, buf, sizeof(Msg *));
	    return sizeof(Msg *);

        default:
            return -1;
    }
}

/*
 * Peek into the PSYNC header of a msg.
 */
static long psync_hdr_peek(hdr, src, len, arg)
VOID *hdr, *arg;
char *src;
long len;
{
    bcopy(src, (char *)hdr, len);
    return 0;
}

/*
 * Strip off the header of a msg.
 */
static long psync_hdr_strip(hdr, src, len, arg)
VOID *hdr, *arg;
char *src;
long len;
{
    bcopy(src, (char *)hdr, len);
    return len;
}

/*
 * Add a header of type PSYNCheader1 or PSYNCheader2 to a msg.  Currently does
 * not have htons calls but can be added.
 */
static void psync_hdr_store(hdr, dst, len, arg)
VOID *hdr, *arg;
char *dst;
long len;
{
    bcopy((char *)hdr, dst, len);
}

/*
 * Store the nodeid of the msg being popped to the protocol above.
 * The nodeid is needed by TOTAL_ORDER.
 * *** UNCONVENTIONAL ***  this is called in the xPop of the hlp.
 */
static void node_hdr_store(hdr, dst, len, arg)
VOID *hdr, *arg;
char *dst;
long len;
{
    ARGstruct *karg = (ARGstruct *)arg;
    NODE      *node;

    xAssert(arg);

    /* `map_tree' is <iphost,mid> --> <node> */
    if (mapResolve((Map)karg->state->map_tree, (VOID *)&karg->mid,
		   (VOID **)&node) == XK_FAILURE)
	Kabort("PSYNC: node not found in Ih for node_hdr_store");
    xTrace1(psyncp, TR_FUNCTIONAL_TRACE,
            "PSYNC: node_hdr_store: storing node %d onto msg", node);
    bcopy((char *)&node, dst, sizeof(NODE *));
}

/*
 * Routine for the thread scheduled to send the msg back to the sending hlp.
 */
static void rec_own_msg(ev, arg)
Event ev;
VOID  *arg;
{
    ARGstruct *karg = (ARGstruct *)arg;

    xTrace1(psyncp, TR_DETAILED,
            "PSYNC: rec_own_msg: msg len = %d before adding nodeid",
            msgLen(karg->msg));
    xIfTrace(psyncp, TR_DETAILED) {
        MSG_SHOW(karg->msg);
    }
    msgPush(karg->msg, node_hdr_store, (VOID *)arg, sizeof(NODE *),
	    (VOID *)arg);
    xTrace1(psyncp, TR_DETAILED,
            "PSYNC: rec_own_msg: msg len = %d after adding nodeid",
            msgLen(karg->msg));
    xDemux(karg->psync_sessn, karg->msg);
    msgDestroy(karg->msg);
    xFree((char *)karg->msg);
    xFree((char *)karg);
}

/*
 * Send reply in response to a RETRANSMIT msg.
 */
static void do_reply(mysessn, depend, nd, hdr2, own)
XObj         mysessn;
NODE         *depend[], *nd;
PSYNCheader2 *hdr2;
int          own;
{
    Map          indegree_map;
    int          qbegin, qend, i, j, index;
    NODE         *send_nodes[350], *node;
    INDEGREE     que[350];
    PSYNCheader2 newhdr2;
    Msg          msg;

    xAssert(xIsSession(mysessn));
    xTrace0(psyncp, 1, "PSYNC: do_reply");

    if (xIsXObj(xGetDown(mysessn,hdr2->sender))) {
        indegree_map = mapCreate(600, sizeof(NODE *));
        compute_indegrees(depend, nd, &indegree_map, que);
        qbegin = 0;
        qend   = 1;
        send_nodes[0] = nd;
        while (qbegin != qend) {
            node = send_nodes[qbegin];
            for (i = 0; i < node->num_prev; i++) {
		/* `indegree_map' is <node> --> <que_index> */
                if (mapResolve(indegree_map, (VOID *)&node->prev[i],
			       (VOID **)&index) != XK_FAILURE) {
                    que[index].indegree--;
                    if (que[index].indegree == 0) {
                        send_nodes[qend] = node->prev[i];
                        qend++;
                    }
                }
            }
            qbegin++;
	}

        for (i = qend - 1; i >= 0; i--) {
            node = send_nodes[i];
            if (own || node != nd) {
		bzero((char *)&newhdr2, sizeof(PSYNCheader2));
                newhdr2.type    = REPLY;
                newhdr2.cid     = hdr2->cid;
                newhdr2.mid     = node->mid;
                newhdr2.sender  = node->sender;
                for (j = 0; j < node->num_prev; j++)
                    newhdr2.depend_list[j] = node->prev[j]->mid;
                newhdr2.num_depend = j;
                xTrace1(psyncp, TR_FUNCTIONAL_TRACE,
                        "PSYNC: do_reply: reply to %d", hdr2->sender);
                msgConstructCopy(&msg, node->msg);
	        dumpHdr2(&newhdr2);
	        xIfTrace(psyncp, TR_DETAILED) {
	            MSG_SHOW(&msg);
	        }
                msgPush(&msg, psync_hdr_store, (VOID *)&newhdr2,
			sizeof(PSYNCheader2), (VOID *)NULL);
	        xTrace0(psyncp, TR_DETAILED,
			"PSYNC: do_reply: sending reply msg with hdr");
	        xIfTrace(psyncp, TR_DETAILED) {
	            MSG_SHOW(&msg);
	        }
                xPush(xGetDown(mysessn,hdr2->sender), &msg);
	        xTrace0(psyncp, TR_ALWAYS,
			"PSYNC: do_reply: after push returns from IP");
	        xIfTrace(psyncp, TR_DETAILED) {
	            MSG_SHOW(&msg);
	        }
		msgDestroy(&msg);
            }
        }
        mapClose(indegree_map);
    }
}

static void do_restart(mysessn, cid, nd, downsessn)
XObj mysessn, downsessn;
CID  *cid;
NODE *nd;
{
    PSYNCheader2 newhdr2;
    int          i;
    Msg          msg;

    Kabort("PSYNC: do_restart - not supported");
    if (downsessn != (XObj)-1) {
	bzero((char *)&newhdr2, sizeof(PSYNCheader2));
        newhdr2.type   = REGULAR;
        newhdr2.cid    = *cid;
        newhdr2.mid    = nd->mid;
        newhdr2.sender = nd->sender;
        for (i = 0; i < nd->num_prev; i++)
            newhdr2.depend_list[i] = nd->prev[i]->mid;
        newhdr2.num_depend = i;
        msgConstructCopy(&msg, nd->msg);
        msgPush(&msg, psync_hdr_store, (VOID *)&newhdr2, sizeof(PSYNCheader2),
		(VOID *)NULL);
        xPush(downsessn, &msg);
	msgDestroy(&msg);
    }
}

/*
 * Add node to the Ih graph.  Put into `depend_list' the current leaves of Ih.
 * Once the message is added check its ancestors for stability, and if stable,
 * free them.  Set status to 0 if message was sucessfully added, -1 if message
 * is already known, 1 if some of the message's parents are missing and a
 * retransmit request is required and 2 if some of message's parents are
 * missing and retransmit request is not required.
 */
static int add_node(pstate, sstate, hdr2, msg, numd, depend_list, push, status)
PSYNCstate    *pstate;
Sstate        *sstate;
PSYNCheader2  *hdr2;
Msg           *msg;
unsigned char *numd;
MID           *depend_list;
int           push, *status;
{
    int  retval, i, j;
    NODE *node, *parent, *prev_nlist[MAX_PART];

    xTrace1(psyncp, TR_FUNCTIONAL_TRACE,
	    "PSYNC: entering add_node, num_depend = %d", hdr2->num_depend);

    semWait(&pstate->Ihgraph);

    /* hdr2->depend_list is list of mids on which this message depends */
    if (hdr2->num_depend == 0) {
        xError(
	    "*** this shouldn't happen; NON-FIRST message with no depends ***");
	*status = 0;
	return 0;
    }

    /* `map_tree' is <iphost,mid> --> <node> */
    if (mapResolve(sstate->map_tree, (VOID *)&hdr2->mid, (VOID **)&node) !=
	    XK_FAILURE) {	/* message itself is already known */
        semSignal(&pstate->Ihgraph);
	xTrace0(psyncp, TR_FUNCTIONAL_TRACE,
		"PSYNC: exiting add_node, message already known");
        *status = -1;
	return (int)node;
    }

    retval = chk_parents(sstate, hdr2->num_depend, hdr2->depend_list,
			 prev_nlist);
    if (retval == 1) {
        for (i = 0, j = sstate->num_leaves; i < j; i++)
	    depend_list[i] = sstate->leaves[i]->mid;
        *numd = j;
        semSignal(&pstate->Ihgraph);
	xTrace0(psyncp, TR_FUNCTIONAL_TRACE,
	      "PSYNC: exiting add_node, parents missing, retransmit required");
	*status = 1;
        return 0;
    }
    if (retval == 2) {
        semSignal(&pstate->Ihgraph);
	xTrace0(psyncp, TR_FUNCTIONAL_TRACE,
	  "PSYNC: exiting add_node, parents missing, retransmit NOT required");
	*status = 2;
        return 0;
    }

    /*
     * Message's parents are known.  prev_nlist[] contains pointers to those
     * nodes of Ih on which the message depends (parent nodes).
     */
    xTrace0(psyncp, TR_DETAILED, "PSYNC: adding node to Ih");
    node = newnode(sstate, &hdr2->mid, hdr2->sender, msg);
    if (push)				/* called from psync_push */
        sstate->lastmsg[hdr2->sender] = node;
    /* Add node to Ih graph. */
    for (i = 0; i < hdr2->num_depend; i++) {
        /* for every node on which the message depends */
        parent = prev_nlist[i];
        node->prev[i] = parent;		/* set reverse link to that node */
    }
    node->num_prev = i;

    /* Bind message id to map_tree. */
    /* `map_tree' is <iphost,mid> --> <node> */
    mapBind(sstate->map_tree, (VOID *)&hdr2->mid, (VOID *)node);
    semSignal(&pstate->Ihgraph);
    xTrace1(psyncp, TR_FUNCTIONAL_TRACE,
	    "PSYNC: exiting add_node, node %d successfully added", (int)node);
    *status = 0;
    return (int)node;
}

/*
 * Put a message on orphan list.
 */
static int add_orphan(sstate, msg)
Sstate *sstate;
Msg    *msg;
{
    int          i;
    PSYNCheader2 hdr2, hdr2b;

    if (sstate->num_orphans >= MAX_ORPHANS) {
        xError("PSYNC: Too many orphans, orphan not added");
        return 0;
    }
    /* Peek into the msg header. */
    msgPop(msg, psync_hdr_peek, (VOID *)&hdr2, sizeof(PSYNCheader2),
	   (VOID *)NULL);
    /* If msg is already in orphan_list return 0, else add msg to the list. */
    for (i = 0; i < sstate->num_orphans; i++) {
        msgPop(sstate->orphan_list[i], psync_hdr_peek, (VOID *)&hdr2b,
	       sizeof(PSYNCheader2), (VOID *)NULL);
        if (SAME_MID(hdr2.mid, hdr2b.mid))
	    return 0;
    }
    sstate->orphan_list[sstate->num_orphans++] = msg;
    return 1;
}

/*
 * Check if any messages on the orphan list have their parents now.  If any,
 * add them to graph and pop them to hlps.
 */
static void chk_orphans(pstate, mysessn)
PSYNCstate *pstate;
XObj       mysessn;
{
    Sstate        *sstate = (Sstate *)mysessn->state;
    Msg           msg;
    PSYNCheader2  hdr2, dummyhdr2;
    NODE          *dummy1[MAX_PART];
    unsigned char num_dummy2;
    MID           dummy2[MAX_PART];
    int           i, status;
    ARGstruct     arg;

    xAssert(xIsSession(mysessn));
    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: entering chk_orphans");

    /*
     * for all the msgs in the orphan_list
     *   peek into the header of the msg to get the mid
     *   see if all the parents of the msg are received
     *   if so
     *     remove the header and add it to the context graph
     *     if cannot add to the context graph
     *       ERROR
     *     else
     *       store the nodeid onto the msg and demux it up
     */
    for (i = 0; i < sstate->num_orphans; ) {	/* walk through orphans list */
        msgConstructCopy(&msg, sstate->orphan_list[i]);
        msgPop(&msg, psync_hdr_peek, (VOID *)&hdr2, sizeof(PSYNCheader2),
	       (VOID *)NULL);
        if (!chk_parents(sstate, hdr2.num_depend, hdr2.depend_list, dummy1)) {
            /* Message has its context now; remove it from orphan list. */
	    msgDestroy(sstate->orphan_list[i]);
	    xFree((char *)sstate->orphan_list[i]);
            sstate->orphan_list[i] = sstate->orphan_list[--sstate->num_orphans];
            msgPop(&msg, psync_hdr_strip, (VOID *)&dummyhdr2,
		   sizeof(PSYNCheader2), (VOID *)NULL);
            (void)add_node(pstate, sstate, &hdr2, &msg, &num_dummy2, dummy2, 0,
			   &status);
            if (status == 0) {
                /* No problems with the msg; send it up. */
                arg.state = sstate;
                arg.mid   = hdr2.mid;
                msgPush(&msg, node_hdr_store, (VOID *)&arg, sizeof(NODE *),
                        (VOID *)&arg);
	        if (!sstate->loopback) {
	            if (update_graph(pstate, sstate, &hdr2.mid))
	                start_stability_check(pstate, sstate, &hdr2.mid);
	        }
	        xDemux(mysessn, &msg);
		msgDestroy(&msg);
	        i = 0;
            }
	    else if (status > 0)
	        Kabort("PSYNC: add_node failed in chk_orphans");
        }
	else
            i++;
    }
    xTrace0(psyncp, TR_FUNCTIONAL_TRACE, "PSYNC: exiting chk_orphans");
}

/*
 * Return 0 if all the parents are known; return 1 if some parents are not
 * known and a retransmit request is required because some of the missing
 * parents are not present in orphan list or because too many orphans have
 * accumulated (probably because of a loss of some reply message); return 2 if
 * some parents are missing but no need for a retransmit request.  Return an
 * array of node pointers of the parents of the message (prev_nlist) if all
 * parents are known.
 */
static int chk_parents(sstate, num_parents, parents, prev_nlist)
Sstate        *sstate;
u_int 	      num_parents;
MID           *parents;
NODE          *prev_nlist[];
{
    int i, missing = 0;
    MID *mid;

    for (i = 0; i < num_parents; i++) {
        /* for every parent message */
        mid = &parents[i];
        /* `map_tree' is <iphost,mid> --> <node> */
        if (mapResolve(sstate->map_tree, (VOID *)mid,
		       (VOID **)&prev_nlist[i]) == XK_FAILURE) {
            xTrace2(psyncp, TR_DETAILED,
                    "message mid->msg_id = %d, mid->host = %s is missing",
                    mid->msg_id, ipHostStr(&mid->host));
            missing = 1;
            if (sstate->num_orphans == 0)
                return 1;
            else if (!chk_orphan_list(sstate, mid))
                return 1;
        }
    }
    if (missing)
        return 2;
    return 0;
}

static int chk_orphan_list(sstate, mid)
Sstate *sstate;
MID    *mid;
{
    int          i;
    PSYNCheader2 hdr2;

    for (i = 0; i < sstate->num_orphans; i++) {
        msgPop(sstate->orphan_list[i], psync_hdr_peek, (VOID *)&hdr2,
               sizeof(PSYNCheader2), (VOID *)NULL);
        if (SAME_MID(hdr2.mid, *mid))
            return TRUE;
    }
    return FALSE;
}
