/*
 * $RCSfile: hash.c,v $
 *
 * Universal Stub Compiler
 *
 * Copyright (c) 1994    Arizona Board of Regents
 *
 * $Revision: 1.1 $
 * $Date: 1994/10/13 17:43:30 $
 */

/****************************************************************/
/*                                                              */
/*  hash.c : routines for the symbol hash table                 */
/*                                                              */
/****************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "hash.h"

/*****************************************************************/
/*                                                               */
/*   assort() : shell sort for pointers                          */
/*                                                               */
/*****************************************************************/


void assort (void **base, int nel, int elsize, int (*cmp)())
{
  int i, j, gap;
  void *tmp, **p1, **p2;

  for (gap = 1; gap <= nel; gap = 3 * gap + 1);

  for (gap /= 3; gap > 0; gap /= 3)
    for (i = gap; i < nel; i++)
      for (j = i - gap; j >= 0; j -= gap) {
	p1 = base + (j);
	p2 = base + ((j + gap));
	if ((*cmp)(p1, p2) <= 0)
	  break;
	tmp = *p1;
	*p1 = *p2;
	*p2 = tmp;
      }
}

void nuketab (Hash_Tab *tabp)
{
  int i;
  for (i = tabp->size; --i >= 0; ) {
	 tabp->table[i] = NULL;
  }
  tabp->numsyms = 0;
}
	 
void *headsym (Hash_Tab *tabp, int n)
{
  return tabp->table[n] ? tabp->table[n] + 1 : NULL;
}

void *walktab (Hash_Tab *tabp, void *isym)
{
  Bucket *sym = (Bucket *)isym;
  sym--;
  if (sym->next)
	 return sym->next + 1;
  else
	 return NULL;
}

void cleantab (Hash_Tab *tabp, int (*test)())
{
  Bucket *sym, **symtab;
  int i;
  for (symtab = tabp->table, i = tabp->size; --i >= 0; symtab++) {
	 for (sym = *symtab; sym; sym = sym->next) {
		if ((*test)(sym + 1))
		  delsym (tabp, sym + 1);
	 }
  }
}

void *newsym (int size)
{
  Bucket *sym;

  if (!(sym = (Bucket *)calloc (size + sizeof (Bucket), 1)))  {
    fprintf (stderr, "calloc error\n");
    exit(1);
  } 
  return (void *) (sym + 1);
}

void freesym (void *sym)
{
  free ((Bucket *)sym - 1);
}

Hash_Tab *maketab (unsigned maxsym, unsigned (*hash_function)(),
						 int (*cmp_function)())
{
  Hash_Tab *p;
  
  if (!maxsym)
    maxsym = 127;

  if (p = (Hash_Tab *)calloc (1, (maxsym * sizeof (Bucket *)) +
			      sizeof (Hash_Tab)) ) {
    p->size = maxsym;
    p->numsyms = 0;
    p->hash = hash_function;
    p->cmp = cmp_function;
  }
  else {
    fprintf (stderr, "calloc error\n");
    exit (1);
  }
  return p;
}

void *addsym (Hash_Tab *tabp, void *isym)
{
  Bucket **p, *tmp;
  Bucket *sym = (Bucket *)isym;

  if (tabp->hash == NULL) {
	 p = &(tabp->table)[0];
	 sym--;
  }
  else {
	 p = &(tabp->table)[(*tabp->hash)(sym--) % tabp->size];
  }
  tmp = *p;
  *p = sym;
  sym->prev = p;
  sym->next = tmp;

  if (tmp)
    tmp->prev = &sym->next;

  tabp->numsyms++;
  return (void *)(sym + 1);
}

void delsym (Hash_Tab *tabp, void *isym)
{
  Bucket *sym = (Bucket *)isym;
  
  if (tabp && sym) {
    --tabp->numsyms;
    --sym;
    
    if (*(sym->prev) = sym->next)
      sym->next->prev = sym->prev;
  }
}

void *findsym (Hash_Tab *tabp, void *sym)
{
  Bucket *p;

  if (!tabp)
    return NULL;
  
  if (tabp->hash == NULL)
	 p = (tabp->table)[0];
  else
	 p = (tabp->table)[(*tabp->hash)(sym) % tabp->size];

  while (p && (*tabp->cmp)(sym, p + 1))
    p = p->next;

  return (void *)(p ? p + 1 : NULL);
}

void *nextsym (Hash_Tab *tabp, void *i_last)
{
  Bucket *last = (Bucket *)i_last;

  for (--last; last->next; last = last->next) {
    if ((tabp->cmp)(last + 1, last->next + 1) == 0)
      return (void *)(last->next + 1);
  }
  return NULL;
}

int (*User_cmp)();

int ptab (Hash_Tab *tabp, void (*print)(), void *param, int sort)
{
  Bucket **outtab, **outp, *sym, **symtab;
  int internal_cmp();
  int i;

  if (!tabp || tabp->size == 0)
    return 1;
  
  if (!sort) {
    for (symtab = tabp->table, i = tabp->size; --i >= 0; symtab++) {
      for (sym = *symtab; sym; sym = sym->next) {
		  (*print)(sym + 1, param);
		}
    }
  }
  else {
    if (!(outtab = (Bucket **)malloc(tabp->numsyms * sizeof (Bucket *))))
      return 0;

    outp = outtab;

    for (symtab = tabp->table, i = tabp->size; --i >= 0; symtab++) {
      for (sym = *symtab; sym; sym = sym->next) {
	if (outp > outtab + tabp->numsyms) {
	  fprintf (stderr, "Internal error [ptab], table overflow\n");
	  exit (1);
	}
	*outp++ = sym;
      }
    }
    
    User_cmp = tabp->cmp;
    assort ((void **)outtab, tabp->numsyms, sizeof (Bucket *), internal_cmp);

    for (outp = outtab, i = tabp->numsyms; --i >= 0; outp++)
      (*print)((*outp) + 1, param);
    
    free (outtab);
  }
  return 1;
}

int internal_cmp (Bucket **p1, Bucket **p2)
{
  return (*User_cmp)(*p1 + 1, *p2 + 1);
}
